from pydae.build_v2 import builder
import numpy as np
import sympy as sym
from sympy.matrices.sparsetools import _doktocsr
from sympy import SparseMatrix

x_1,x_2,x_3 = sym.symbols('x_1,x_2,x_3', real=True)
y_1,y_2,y_3,y_4 = sym.symbols('y_1,y_2,y_3,y_4', real=True)
u_1,u_2 = sym.symbols('u_1,u_2', real=True)

p_f1,p_f2,p_f3,p_f4 = sym.symbols('p_f1,p_f2,p_f3,p_f4', real=True)
p_g1,p_g2,p_g3,p_g4 = sym.symbols('p_g1,p_g2,p_g3,p_g4', real=True)
p_u1,p_u2,p_u3,p_u4 = sym.symbols('p_u1,p_u2,p_u3,p_u4', real=True)
p_h1,p_h2,p_h3,p_h4 = sym.symbols('p_h1,p_h2,p_h3,p_h4', real=True)




x = sym.Matrix([[x_1],   
                [x_2],   
                [x_3]])

y = sym.Matrix([[y_1],   
                [y_2],   
                [y_3],   
                [y_4]])

u = sym.Matrix([[u_1],   
                [u_2]])


f = sym.Matrix([[-p_f1*x_1 + p_f4*y_1 + p_u1*u_1 + x_3*y_1], 
                [-p_f2*x_3], 
                [-p_f3*x_2 + u_2*y_1]])
                
g = sym.Matrix([[p_g1*y_2 + p_u2*u_1], 
                [p_g2*y_1 + x_2*y_1], 
                [p_g3*y_3 + x_3*y_2], 
                [p_u3*u_2 + u_1*y_4 + x_2*y_2]])

h = sym.Matrix([[p_h1*x_1 + p_h2*y_1 + p_h3*y_4 + p_h4*u_1], 
                [u_2*x_1 + x_1*y_2 + x_2*y_1]])

z = h

F_x = f.jacobian(x)
F_y = f.jacobian(y)
F_u = f.jacobian(u)

G_x = g.jacobian(x)
G_y = g.jacobian(y)
G_u = g.jacobian(u)

H_x = h.jacobian(x)
H_y = h.jacobian(y)
H_u = h.jacobian(u)

jac_ini_sym = sym.Matrix([[F_x,F_y],[G_x,G_y]]) 
#print(jac_ini_sym)
jac_ini_data,jac_ini_indices,jac_ini_indptr,jac_ini_shape = _doktocsr(SparseMatrix(jac_ini_sym))

jac_ini_sym = sym.Matrix([[F_x,F_y],[G_x,G_y]]) 
#print(jac_ini_sym)
jac_ini_data,jac_ini_indices,jac_ini_indptr,jac_ini_shape = _doktocsr(SparseMatrix(jac_ini_sym))

N_x = F_x.shape[0]
eye = sym.eye(N_x, real=True)
Dt = sym.Symbol('Dt',real=True)
jac_trap = sym.Matrix([[eye - 0.5*Dt*F_x, -0.5*Dt*F_y],[G_x,G_y]])    
print('jac_trap = ',jac_trap)
jac_trap_data,jac_trap_indices,jac_trap_indptr,jac_trap_shape = _doktocsr(SparseMatrix(jac_trap_sym))

print('F_x = ',F_x)
print('F_y = ',F_y)
print('F_u = ',F_u)

print('G_x = ',G_x)
print('G_y = ',G_y)
print('G_u = ',G_u)

print('H_x = ',H_x)
print('H_y = ',H_y)
print('H_u = ',H_u)


p_f1,p_f2,p_f3,p_f4= 1.0,2.0,3.0,4.0
p_g1,p_g2,p_g3,p_g4= 1.0,2.0,3.0,4.0
p_u1,p_u2,p_u3,p_u4= 1.0,2.0,3.0,4.0
p_h1,p_h2,p_h3,p_h4= 1.0,2.0,3.0,4.0

params_dict = {
'p_f1':p_f1,'p_f2':p_f2,'p_f3':p_f3,'p_f4':p_f4,
'p_g1':p_g1,'p_g2':p_g2,'p_g3':p_g3,'p_g4':p_g4,
'p_u1':p_u1,'p_u2':p_u2,'p_u3':p_u3,'p_u4':p_u4,
'p_h1':p_h1,'p_h2':p_h2,'p_h3':p_h3,'p_h4':p_h4,
}

u_1,u_2 = 1.0,2.0

u_dict = {
'u_1':1.0,'u_2':2.0
}

h_dict ={
'z_1':z[0],'z_2':z[1]
} 

sys_dict = {'name':'dae1','uz_jacs':True,
            'params_dict':params_dict,
            'f_list':list(f),
            'g_list':list(g),
            'x_list':list(x),
            'y_ini_list':list(y),
            'y_run_list':list(y),
            'u_run_dict':u_dict,
            'u_ini_dict':u_dict,
            'h_dict':h_dict}

b = builder(sys_dict)
b.dict2system()
b.functions()
b.jacobians()
b.cwrite()
b.compile()
b.template()

import dae1

model = dae1.model()
model.ini({},1)
#model.report_x()
#model.report_y()
#model.report_u()
#model.report_z()
#model.report_params()
#model.jac_ini

x_1,x_2,x_3 = model.get_mvalue(['x_1','x_2','x_3'])
y_1,y_2,y_3,y_4 = model.get_mvalue(['y_1','y_2','y_3','y_4'])

x = np.array([[x_1],   
              [x_2],   
              [x_3]]).reshape(3,)

y = np.array([[y_1],   
              [y_2],   
              [y_3],   
              [y_4]]).reshape(4,)

u = np.array([[u_1],   
              [u_2]]).reshape(2,)


p = np.array([p_f1,p_f2,p_f3,p_f4,
              p_g1,p_g2,p_g3,p_g4,
              p_u1,p_u2,p_u3,p_u4,
              p_h1,p_h2,p_h3,p_h4]).reshape(16,)

Dt = 1e-3

F_x =  np.array([[-p_f1, 0, y_1], [0, 0, -p_f2], [0, -p_f3, 0]])
F_y =  np.array([[p_f4 + x_3, 0, 0, 0], [0, 0, 0, 0], [u_2, 0, 0, 0]])
F_u =  np.array([[p_u1, 0], [0, 0], [0, y_1]])
G_x =  np.array([[0, 0, 0], [0, y_1, 0], [0, 0, y_2], [0, y_2, 0]])
G_y =  np.array([[0, p_g1, 0, 0], [p_g2 + x_2, 0, 0, 0], [0, x_3, p_g3, 0], [0, x_2, 0, u_1]])
G_u =  np.array([[p_u2, 0], [0, 0], [0, 0], [y_4, p_u3]])

jac_ini = np.array([[-p_f1, 0, y_1, p_f4 + x_3, 0, 0, 0], 
                    [0, 0, -p_f2, 0, 0, 0, 0], 
                    [0, -p_f3, 0, u_2, 0, 0, 0], 
                    [0, 0, 0, 0, p_g1, 0, 0], 
                    [0, y_1, 0, p_g2 + x_2, 0, 0, 0], 
                    [0, 0, y_2, 0, x_3, p_g3, 0], 
                    [0, y_2, 0, 0, x_2, 0, u_1]])


assert (np.max(np.abs(jac_ini-model.jac_ini))) < 1e-16

from dae1 import de_jac_ini_eval,sp_jac_ini_eval

N_x = F_x.shape[0]
N_y = G_y.shape[0]
N_xy = N_x+N_y
de_jac_ini = np.zeros((N_xy,N_xy))
de_jac_ini_eval(de_jac_ini,x,y,u,p,Dt)
#print(np.max(np.abs(jac_ini-de_jac_ini)))

from scipy.sparse import csr_array
from scipy.sparse import csr_matrix

sp_jac_ini = csr_matrix((np.zeros(len(jac_ini_data)),jac_ini_indices,jac_ini_indptr),shape=jac_ini_shape)
sp_jac_ini_eval(sp_jac_ini.data,x,y,u,p,Dt)

assert (np.max(np.abs(jac_ini-sp_jac_ini))) < 1e-16
#print(de_jac_ini)
#print(sp_jac_ini.toarray())
#print(sp_jac_ini.indices)
#print(sp_jac_ini.indptr)
#print(sp_jac_ini.nnz)
#
#print(jac_ini_indices,jac_ini_indptr)

jac_trap =  np.array([[0.5*Dt*p_f1 + 1, 0, -0.5*Dt*y_1, -0.5*Dt*(p_f4 + x_3), 0, 0, 0], 
                      [0, 1, 0.5*Dt*p_f2, 0, 0, 0, 0], 
                      [0, 0.5*Dt*p_f3, 1, -0.5*Dt*u_2, 0, 0, 0], 
                      [0, 0, 0, 0, p_g1, 0, 0], 
                      [0, y_1, 0, p_g2 + x_2, 0, 0, 0], 
                      [0, 0, y_2, 0, x_3, p_g3, 0], 
                      [0, y_2, 0, 0, x_2, 0, u_1]])


# -*- coding: utf-8 -*-
"""
Created on Wed Dec 15 23:38:58 2021

@author: jmmau
"""

import pytest
import numpy as np
from pydae.bmapu import bmapu_builder
import pydae.ssa as ssa
 
 
def test_vsc_l_vsg_ll_builder():

    zeta = 0.1 # 1.0/np.sqrt(2) 
    H_v = 4.0 
    WB = 2 *np.pi* 50
    R_v = 0.0
    X_v = 0.3

    Lt = X_v 
    P_max = 1/Lt
    fn = np.sqrt(WB*P_max/(2*H_v))/(2*np.pi)

    Dp = 0
    K_i = (2*np.pi*fn)**2/(WB*P_max)
    K_g = Dp*K_i
    K_p = (2*zeta*2*np.pi*fn - K_g)/(WB*P_max)

    T_q = 1.0/(2*np.pi*10/2)
    K_q = (1.0 - 0.0)/(1.05 - 0.95)
    K_i_q = 1e-6

    theta_red = 3.0/180*np.pi
    V = 1.0
    p_ref = 0.9
    q_ref = 0.434616
    v_ref = 1.0
    T_q = 1.0/(2*np.pi*10/2)


    data = {
    "system":{"name":"smib_vsc_l","S_base":100e6, "K_p_agc":0.0,"K_i_agc":0.0,"K_xif":0.01},       
    "buses":[{"name":"1", "P_W":0.0,"Q_var":0.0,"U_kV":20.0},
            {"name":"2", "P_W":0.0,"Q_var":0.0,"U_kV":20.0}
            ],
    "lines":[{"bus_j":"1", "bus_k":"2", "X_pu":0.05,"R_pu":0.01,"Bs_pu":1e-6,"S_mva":100.0}],
    "vscs": [{"bus":"1","type":"vsc_l","S_n":10e6,"F_n":50.0,"X_s":0.05,"R_s":0.005,
          "ctrl":{"type":"leon_vsg_ll","K_delta":0.0,"F_n":50.0,
                  "R_v":R_v,"X_v":X_v,"K_p":K_p,"K_i":K_i,"K_g":K_g,"K_q":K_q,
                  "T_q":T_q,"K_p_v":1e-6,"K_i_v":1e-6}}],
    "genapes":[
        {"bus":"2","S_n":100e6,"F_n":50.0,"X_v":0.001,"R_v":0.0,"K_delta":0.001,"K_alpha":1e-6}]
    }
    
    grid = bmapu_builder.bmapu(data)
    #grid.checker()
    grid.uz_jacs = True
    grid.verbose = False
    grid.build('smib_vsc_l_vsg_ll')


def test_vsc_l_vsg_ll():

    import smib_vsc_l_vsg_ll

    model = smib_vsc_l_vsg_ll.model()
    zeta = 0.2 # 1.0/np.sqrt(2) 
    H_v = 4.0 
    WB = 2 *np.pi* 50;
    R_v = 0.0
    X_v = 0.3

    Lt = X_v 
    P_max = 1/Lt
    fn = np.sqrt(WB*P_max/(2*H_v))/(2*np.pi)

    Dp = 0;
    K_i = (2*np.pi*fn)**2/(WB*P_max);
    K_g = Dp*K_i;
    K_p = (2*zeta*2*np.pi*fn - K_g)/(WB*P_max);

    T_q = 1.0/(2*np.pi*10/2)
    K_q = (1.0 - 0.0)/(1.05 - 0.95)
    K_i_q = 1e-6

    theta_red = 3.0/180*np.pi
    V = 1.0
    p_ref = 0.9
    q_ref = 0.434616
    v_ref = 1.0
    T_q = 1.0/(2*np.pi*10/2)

    p_l_1 = 0.5
    v_ref_1 = 1.01
    X_s = 0.1
    R_s = 0.0
    params = {"S_n_1":10e6,'p_l_1':p_l_1,'v_ref_1':v_ref_1,'X_s_1':X_s,'R_s_1':R_s,
            'b_1_2':-2,
            'X_v_1': X_v-X_s, 'R_v_1':R_v,
            'T_q_1':T_q,'K_q_1':K_q,
            'K_i_1':K_i,'K_g_1':K_g,'K_p_1':K_p}

    model.ini(params,'xy_0.json')



    p_s_1 = model.get_value('p_s_1')
    q_s_1 = model.get_value('q_s_1')
    V_1 = model.get_value('V_1')

    # Check active and voltage references are achived
    tol = 1e-8
    assert np.abs(p_l_1 - p_s_1) < tol  # p_s
    assert np.abs(v_ref_1 - V_1) < tol  # V_s

    # check dq magnitudes are correct
    delta = model.get_value('delta_1')
    q_s = model.get_value('q_s_1')
    V_s = model.get_value('V_1')
    i_sr = model.get_value('i_sr_1')  
    i_si = model.get_value('i_si_1')  

    #v_sd = v_si * np.cos(delta) - v_sr * np.sin(delta)   
    #v_sq = v_si * np.sin(delta) + v_sr * np.cos(delta)
    i_sd_exp = i_si * np.cos(delta) - i_sr * np.sin(delta)   
    i_sq_exp = i_si * np.sin(delta) + i_sr * np.cos(delta)

    i_sd = model.get_value('i_sd_1')  
    i_sq = model.get_value('i_sq_1') 

    assert np.abs(i_sd_exp - i_sd) < tol  # i_sd
    assert np.abs(i_sq_exp - i_sq) < tol  # i_sq

    # check virtual impedance
    e_vd = 0.0
    e_vq = model.get_value('e_vq_1')
    R_v = model.get_value('R_v_1')
    X_v = model.get_value('X_v_1')
    v_td_exp = e_vd - R_v*i_sd_exp - X_v*i_sq_exp   
    v_tq_exp = e_vq - R_v*i_sq_exp + X_v*i_sd_exp   

    v_td = model.get_value('v_td_ref_1')
    v_tq = model.get_value('v_tq_ref_1')

    assert np.abs(v_td_exp - v_td) < tol  # v_td
    assert np.abs(v_tq_exp - v_tq) < tol  # v_tq

    # check control output
    v_dc = model.get_value('v_dc_1')
    m_exp = np.sqrt(v_td_exp**2 + v_tq_exp**2)/v_dc

    v_ti_exp =  v_td_exp * np.cos(delta) + v_tq_exp * np.sin(delta)   
    v_tr_exp = -v_td_exp * np.sin(delta) + v_tq_exp * np.cos(delta) 

    theta_t_exp = np.arctan2(v_ti_exp,v_tr_exp)
    theta_t = model.get_value('theta_t_1')

    assert np.abs(theta_t_exp - theta_t) < tol  # v_tq


    ssa.A_eval(model)
    damp = ssa.damp_report(model)
    assert  damp.sort_values('Damp').round(2)['Damp'][0] > 0.0
    assert  np.abs(damp.sort_values('Damp').round(2)['Damp'][0] - zeta) < 1e-2

if __name__ == "__main__":
    
    #test_pendulum_builder()
    pass 
    #test_leon_vsg_ll()


import numpy as np
import numba
import scipy.optimize as sopt
import scipy.sparse as sspa
from scipy.sparse.linalg import spsolve,spilu,splu
from numba import cuda
import cffi
import numba.core.typing.cffi_utils as cffi_support

ffi = cffi.FFI()

import proyecto_cffi as jacs

cffi_support.register_module(jacs)
f_ini_eval = jacs.lib.f_ini_eval
g_ini_eval = jacs.lib.g_ini_eval
f_run_eval = jacs.lib.f_run_eval
g_run_eval = jacs.lib.g_run_eval
h_eval  = jacs.lib.h_eval

de_jac_ini_xy_eval = jacs.lib.de_jac_ini_xy_eval
de_jac_ini_up_eval = jacs.lib.de_jac_ini_up_eval
de_jac_ini_num_eval = jacs.lib.de_jac_ini_num_eval

sp_jac_ini_xy_eval = jacs.lib.sp_jac_ini_xy_eval
sp_jac_ini_up_eval = jacs.lib.sp_jac_ini_up_eval
sp_jac_ini_num_eval = jacs.lib.sp_jac_ini_num_eval

de_jac_trap_xy_eval= jacs.lib.de_jac_trap_xy_eval            
de_jac_trap_up_eval= jacs.lib.de_jac_trap_up_eval        
de_jac_trap_num_eval= jacs.lib.de_jac_trap_num_eval

sp_jac_trap_xy_eval= jacs.lib.sp_jac_trap_xy_eval            
sp_jac_trap_up_eval= jacs.lib.sp_jac_trap_up_eval        
sp_jac_trap_num_eval= jacs.lib.sp_jac_trap_num_eval

import json

sin = np.sin
cos = np.cos
atan2 = np.arctan2
sqrt = np.sqrt 
sign = np.sign 
exp = np.exp


class proyecto_class: 

    def __init__(self): 

        self.t_end = 10.000000 
        self.Dt = 0.0010000 
        self.decimation = 10.000000 
        self.itol = 1e-6 
        self.Dt_max = 0.001000 
        self.Dt_min = 0.001000 
        self.solvern = 5 
        self.imax = 100 
        self.N_x = 8
        self.N_y = 20 
        self.N_z = 17 
        self.N_store = 10000 
        self.params_list = ['S_base', 'g_GRI_POI', 'b_GRI_POI', 'g_POI_PMV', 'b_POI_PMV', 'g_PMV_GR1', 'b_PMV_GR1', 'g_GR1_GR2', 'b_GR1_GR2', 'g_PMV_GR3', 'b_PMV_GR3', 'g_GR3_GR4', 'b_GR3_GR4', 'U_GRI_n', 'U_POI_n', 'U_PMV_n', 'U_GR1_n', 'U_GR2_n', 'U_GR3_n', 'U_GR4_n', 'S_n_GRI', 'X_d_GRI', 'X1d_GRI', 'T1d0_GRI', 'X_q_GRI', 'X1q_GRI', 'T1q0_GRI', 'R_a_GRI', 'X_l_GRI', 'H_GRI', 'D_GRI', 'Omega_b_GRI', 'omega_s_GRI', 'K_a_GRI', 'T_r_GRI', 'v_pss_GRI', 'Droop_GRI', 'T_m_GRI', 'K_sec_GRI', 'K_delta_GRI', 'v_ref_GRI', 'e_bess_0'] 
        self.params_values_list  = [100000000.0, 1.4986238532110094, -4.995412844036698, 2.941176470588235, -11.76470588235294, 24.742268041237114, -10.996563573883162, 24.742268041237114, -10.996563573883162, 24.742268041237114, -10.996563573883162, 24.742268041237114, -10.996563573883162, 66000.0, 66000.0, 20000.0, 20000.0, 20000.0, 20000.0, 20000.0, 100000000.0, 1.81, 0.3, 8.0, 1.76, 0.65, 1.0, 0.003, 0.05, 6.0, 1.0, 314.1592653589793, 1.0, 100, 0.1, 0.0, 0.05, 5.0, 0.001, 0.01, 1.0, 10] 
        self.inputs_ini_list = ['P_GRI', 'Q_GRI', 'P_POI', 'Q_POI', 'P_PMV', 'Q_PMV', 'P_GR1', 'Q_GR1', 'P_GR2', 'Q_GR2', 'P_GR3', 'Q_GR3', 'P_GR4', 'Q_GR4'] 
        self.inputs_ini_values_list  = [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1000000.0, 0.0, 1000000.0, 0.0, 1000000.0, 0.0, 1000000.0, 0.0] 
        self.inputs_run_list = ['P_GRI', 'Q_GRI', 'P_POI', 'Q_POI', 'P_PMV', 'Q_PMV', 'P_GR1', 'Q_GR1', 'P_GR2', 'Q_GR2', 'P_GR3', 'Q_GR3', 'P_GR4', 'Q_GR4'] 
        self.inputs_run_values_list = [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1000000.0, 0.0, 1000000.0, 0.0, 1000000.0, 0.0, 1000000.0, 0.0] 
        self.outputs_list = ['V_GRI', 'V_POI', 'V_PMV', 'V_GR1', 'V_GR2', 'V_GR3', 'V_GR4', 'P_GR1', 'Q_GR1', 'P_GR2', 'Q_GR2', 'P_GR3', 'Q_GR3', 'P_GR4', 'Q_GR4', 'P_PMV', 'Q_PMV'] 
        self.x_list = ['delta_GRI', 'omega_GRI', 'e1q_GRI', 'e1d_GRI', 'v_c_GRI', 'p_m_GRI', 'xi_m_GRI', 'e_bess'] 
        self.y_run_list = ['V_GRI', 'theta_GRI', 'V_POI', 'theta_POI', 'V_PMV', 'theta_PMV', 'V_GR1', 'theta_GR1', 'V_GR2', 'theta_GR2', 'V_GR3', 'theta_GR3', 'V_GR4', 'theta_GR4', 'i_d_GRI', 'i_q_GRI', 'P_GRI_1', 'Q_GRI_1', 'v_f_GRI', 'p_m_ref_GRI'] 
        self.xy_list = self.x_list + self.y_run_list 
        self.y_ini_list = ['V_GRI', 'theta_GRI', 'V_POI', 'theta_POI', 'V_PMV', 'theta_PMV', 'V_GR1', 'theta_GR1', 'V_GR2', 'theta_GR2', 'V_GR3', 'theta_GR3', 'V_GR4', 'theta_GR4', 'i_d_GRI', 'i_q_GRI', 'P_GRI_1', 'Q_GRI_1', 'v_f_GRI', 'p_m_ref_GRI'] 
        self.xy_ini_list = self.x_list + self.y_ini_list 
        self.t = 0.0
        self.it = 0
        self.it_store = 0
        self.xy_prev = np.zeros((self.N_x+self.N_y,1))
        self.initialization_tol = 1e-6
        self.N_u = len(self.inputs_run_list) 
        self.sopt_root_method='hybr'
        self.sopt_root_jac=True
        self.u_ini_list = self.inputs_ini_list
        self.u_ini_values_list = self.inputs_ini_values_list
        self.u_run_list = self.inputs_run_list
        self.u_run_values_list = self.inputs_run_values_list
        self.N_u = len(self.u_run_list)
        self.u_ini = np.array(self.inputs_ini_values_list)
        self.p = np.array(self.params_values_list)
        self.xy_0 = np.zeros((self.N_x+self.N_y,))
        self.xy = np.zeros((self.N_x+self.N_y,))
        self.z = np.zeros((self.N_z,))
        
        self.jac_ini = np.zeros((self.N_x+self.N_y,self.N_x+self.N_y))
        self.jac_run = np.zeros((self.N_x+self.N_y,self.N_x+self.N_y))
        self.jac_trap = np.zeros((self.N_x+self.N_y,self.N_x+self.N_y))
        
        self.yini2urun = list(set(self.u_run_list).intersection(set(self.y_ini_list)))
        self.uini2yrun = list(set(self.y_run_list).intersection(set(self.u_ini_list)))
        self.Time = np.zeros(self.N_store)
        self.X = np.zeros((self.N_store,self.N_x))
        self.Y = np.zeros((self.N_store,self.N_y))
        self.Z = np.zeros((self.N_store,self.N_z))
        self.iters = np.zeros(self.N_store) 
        self.u_run = np.array(self.u_run_values_list,dtype=np.float64)
        
        self.sp_jac_trap_ia, self.sp_jac_trap_ja, self.sp_jac_trap_nia, self.sp_jac_trap_nja = sp_jac_trap_vectors()
        data = np.array(self.sp_jac_trap_ia,dtype=np.float64)
        self.sp_jac_trap = sspa.csr_matrix((data, self.sp_jac_trap_ia, self.sp_jac_trap_ja), shape=(self.sp_jac_trap_nia,self.sp_jac_trap_nja))

        self.J_run_d = np.array(self.sp_jac_trap_ia)*0.0
        self.J_run_i = np.array(self.sp_jac_trap_ia)
        self.J_run_p = np.array(self.sp_jac_trap_ja)
        
        self.J_trap_d = np.array(self.sp_jac_trap_ia)*0.0
        self.J_trap_i = np.array(self.sp_jac_trap_ia)
        self.J_trap_p = np.array(self.sp_jac_trap_ja)
        
        self.sp_jac_ini_ia, self.sp_jac_ini_ja, self.sp_jac_ini_nia, self.sp_jac_ini_nja = sp_jac_ini_vectors()
        data = np.array(self.sp_jac_ini_ia,dtype=np.float64)
        self.sp_jac_ini = sspa.csr_matrix((data, self.sp_jac_ini_ia, self.sp_jac_ini_ja), shape=(self.sp_jac_ini_nia,self.sp_jac_ini_nja))

        self.J_ini_d = np.array(self.sp_jac_ini_ia)*0.0
        self.J_ini_i = np.array(self.sp_jac_ini_ia)
        self.J_ini_p = np.array(self.sp_jac_ini_ja)
        

        
        self.max_it,self.itol,self.store = 50,1e-8,1 
        self.lmax_it,self.ltol,self.ldamp=50,1e-8,1.1
        self.mode = 0 

        self.lmax_it_ini,self.ltol_ini,self.ldamp_ini=50,1e-8,1.1

        self.fill_factor_ini,self.drop_tol_ini,self.drop_rule_ini = 10,0.001,'column'       
        self.fill_factor_run,self.drop_tol_run,self.drop_rule_run = 10,0.001,'column' 
        
        # numerical elements of jacobians computing:
        x = self.xy[:self.N_x]
        y = self.xy[self.N_x:]
        
        de_jac_ini_eval(self.jac_ini,x,y,self.u_ini,self.p,self.Dt)
        de_jac_trap_eval(self.jac_trap,x,y,self.u_run,self.p,self.Dt)
   
        sp_jac_ini_eval(self.J_ini_d,x,y,self.u_ini,self.p,self.Dt)
        sp_jac_trap_eval(self.J_trap_d,x,y,self.u_run,self.p,self.Dt)
   


        
    def update(self):

        self.Time = np.zeros(self.N_store)
        self.X = np.zeros((self.N_store,self.N_x))
        self.Y = np.zeros((self.N_store,self.N_y))
        self.Z = np.zeros((self.N_store,self.N_z))
        self.iters = np.zeros(self.N_store)
        
    def ss_ini(self):

        xy_ini,it = sstate(self.xy_0,self.u_ini,self.p,self.jac_ini,self.N_x,self.N_y)
        self.xy_ini = xy_ini
        self.N_iters = it
        
        return xy_ini
    
    # def ini(self,up_dict,xy_0={}):

    #     for item in up_dict:
    #         self.set_value(item,up_dict[item])
            
    #     self.xy_ini = self.ss_ini()
    #     self.ini2run()
    #     jac_run_ss_eval_xy(self.jac_run,self.x,self.y_run,self.u_run,self.p)
    #     jac_run_ss_eval_up(self.jac_run,self.x,self.y_run,self.u_run,self.p)
        
        
        
    
    def run(self,t_end,up_dict):
        for item in up_dict:
            self.set_value(item,up_dict[item])
            
        t = self.t
        p = self.p
        it = self.it
        it_store = self.it_store
        xy = self.xy
        u = self.u_run
        
        t,it,it_store,xy = daesolver(t,t_end,it,it_store,xy,u,p,
                                  self.jac_trap,
                                  self.Time,
                                  self.X,
                                  self.Y,
                                  self.Z,
                                  self.iters,
                                  self.Dt,
                                  self.N_x,
                                  self.N_y,
                                  self.N_z,
                                  self.decimation,
                                  max_it=50,itol=1e-8,store=1)
        
        self.t = t
        self.it = it
        self.it_store = it_store
        self.xy = xy
 
    def runsp(self,t_end,up_dict):
        for item in up_dict:
            self.set_value(item,up_dict[item])
            
        t = self.t
        p = self.p
        it = self.it
        it_store = self.it_store
        xy = self.xy
        u = self.u_run
        
        t,it,it_store,xy = daesolver_sp(t,t_end,it,it_store,xy,u,p,
                                  self.sp_jac_trap,
                                  self.Time,
                                  self.X,
                                  self.Y,
                                  self.Z,
                                  self.iters,
                                  self.Dt,
                                  self.N_x,
                                  self.N_y,
                                  self.N_z,
                                  self.decimation,
                                  max_it=50,itol=1e-8,store=1)
        
        self.t = t
        self.it = it
        self.it_store = it_store
        self.xy = xy
        
    def post(self):
        
        self.Time = self.Time[:self.it_store]
        self.X = self.X[:self.it_store]
        self.Y = self.Y[:self.it_store]
        self.Z = self.Z[:self.it_store]
        
    def ini2run(self):
        
        ## y_ini to y_run
        self.y_ini = self.xy_ini[self.N_x:]
        self.y_run = np.copy(self.y_ini)
        self.u_run = np.copy(self.u_ini)
        
        ## y_ini to u_run
        for item in self.yini2urun:
            self.u_run[self.u_run_list.index(item)] = self.y_ini[self.y_ini_list.index(item)]
                
        ## u_ini to y_run
        for item in self.uini2yrun:
            self.y_run[self.y_run_list.index(item)] = self.u_ini[self.u_ini_list.index(item)]
            
        
        self.x = self.xy_ini[:self.N_x]
        self.xy[:self.N_x] = self.x
        self.xy[self.N_x:] = self.y_run
        c_h_eval(self.z,self.x,self.y_run,self.u_ini,self.p,self.Dt)
        

        
    def get_value(self,name):
        
        if name in self.inputs_run_list:
            value = self.u_run[self.inputs_run_list.index(name)]
            return value
            
        if name in self.x_list:
            idx = self.x_list.index(name)
            value = self.xy[idx]
            return value
            
        if name in self.y_run_list:
            idy = self.y_run_list.index(name)
            value = self.xy[self.N_x+idy]
            return value
        
        if name in self.params_list:
            idp = self.params_list.index(name)
            value = self.p[idp]
            return value
            
        if name in self.outputs_list:
            idz = self.outputs_list.index(name)
            value = self.z[idz]
            return value

    def get_values(self,name):
        if name in self.x_list:
            values = self.X[:,self.x_list.index(name)]
        if name in self.y_run_list:
            values = self.Y[:,self.y_run_list.index(name)]
        if name in self.outputs_list:
            values = self.Z[:,self.outputs_list.index(name)]
                        
        return values

    def get_mvalue(self,names):
        '''

        Parameters
        ----------
        names : list
            list of variables names to return each value.

        Returns
        -------
        mvalue : TYPE
            list of value of each variable.

        '''
        mvalue = []
        for name in names:
            mvalue += [self.get_value(name)]
                        
        return mvalue
    
    def set_value(self,name_,value):
        if name_ in self.inputs_ini_list:
            self.u_ini[self.inputs_ini_list.index(name_)] = value
            #return
        if name_ in self.inputs_run_list:
            self.u_run[self.inputs_run_list.index(name_)] = value
            return
        elif name_ in self.params_list:
            self.p[self.params_list.index(name_)] = value
            return
        else:
            print(f'Input or parameter {name_} not found.')
 
    def report_x(self,value_format='5.2f'):
        for item in self.x_list:
            print(f'{item:5s} = {self.get_value(item):5.2f}')

    def report_y(self,value_format='5.2f'):
        for item in self.y_run_list:
            print(f'{item:5s} = {self.get_value(item):5.2f}')
            
    def report_u(self,value_format='5.2f'):
        for item in self.inputs_run_list:
            print(f'{item:5s} = {self.get_value(item):5.2f}')

    def report_z(self,value_format='5.2f'):
        for item in self.outputs_list:
            print(f'{item:5s} = {self.get_value(item):5.2f}')

    def report_params(self,value_format='5.2f'):
        for item in self.params_list:
            print(f'{item:5s} = {self.get_value(item):5.2f}')
            
    def ini(self,up_dict,xy_0={}):
        
        self.it = 0
        self.it_store = 0
        self.t = 0.0
    
        for item in up_dict:
            self.set_value(item,up_dict[item])
            
        if type(xy_0) == dict:
            xy_0_dict = xy_0
            self.dict2xy0(xy_0_dict)
            
        if type(xy_0) == str:
            if xy_0 == 'eval':
                N_x = self.N_x
                self.xy_0_new = np.copy(self.xy_0)*0
                xy0_eval(self.xy_0_new[:N_x],self.xy_0_new[N_x:],self.u_ini,self.p)
                self.xy_0_evaluated = np.copy(self.xy_0_new)
                self.xy_0 = np.copy(self.xy_0_new)
            else:
                self.load_xy_0(file_name = xy_0)
                
        if type(xy_0) == float or type(xy_0) == int:
            self.xy_0 = np.ones(self.N_x+self.N_y,dtype=np.float64)*xy_0


        self.xy_ini = self.ss_ini()
        self.ini2run()
        #jac_run_ss_eval_xy(self.jac_run,self.x,self.y_run,self.u_run,self.p)
        #jac_run_ss_eval_up(self.jac_run,self.x,self.y_run,self.u_run,self.p)
    
    def dict2xy0(self,xy_0_dict):
    
        for item in xy_0_dict:
            if item in self.x_list:
                self.xy_0[self.x_list.index(item)] = xy_0_dict[item]
            if item in self.y_ini_list:
                self.xy_0[self.y_ini_list.index(item) + self.N_x] = xy_0_dict[item]
        
    
    def save_xy_0(self,file_name = 'xy_0.json'):
        xy_0_dict = {}
        for item in self.x_list:
            xy_0_dict.update({item:self.get_value(item)})
        for item in self.y_ini_list:
            xy_0_dict.update({item:self.get_value(item)})
    
        xy_0_str = json.dumps(xy_0_dict, indent=4)
        with open(file_name,'w') as fobj:
            fobj.write(xy_0_str)
    
    def load_xy_0(self,file_name = 'xy_0.json'):
        with open(file_name) as fobj:
            xy_0_str = fobj.read()
        xy_0_dict = json.loads(xy_0_str)
    
        for item in xy_0_dict:
            if item in self.x_list:
                self.xy_0[self.x_list.index(item)] = xy_0_dict[item]
            if item in self.y_ini_list:
                self.xy_0[self.y_ini_list.index(item)+self.N_x] = xy_0_dict[item]            

    def load_params(self,data_input):

        if type(data_input) == str:
            json_file = data_input
            self.json_file = json_file
            self.json_data = open(json_file).read().replace("'",'"')
            data = json.loads(self.json_data)
        elif type(data_input) == dict:
            data = data_input

        self.data = data
        for item in self.data:
            self.struct[0][item] = self.data[item]
            if item in self.params_list:
                self.params_values_list[self.params_list.index(item)] = self.data[item]
            elif item in self.inputs_ini_list:
                self.inputs_ini_values_list[self.inputs_ini_list.index(item)] = self.data[item]
            elif item in self.inputs_run_list:
                self.inputs_run_values_list[self.inputs_run_list.index(item)] = self.data[item]
            else: 
                print(f'parameter or input {item} not found')

    def save_params(self,file_name = 'parameters.json'):
        params_dict = {}
        for item in self.params_list:
            params_dict.update({item:self.get_value(item)})

        params_dict_str = json.dumps(params_dict, indent=4)
        with open(file_name,'w') as fobj:
            fobj.write(params_dict_str)

    def save_inputs_ini(self,file_name = 'inputs_ini.json'):
        inputs_ini_dict = {}
        for item in self.inputs_ini_list:
            inputs_ini_dict.update({item:self.get_value(item)})

        inputs_ini_dict_str = json.dumps(inputs_ini_dict, indent=4)
        with open(file_name,'w') as fobj:
            fobj.write(inputs_ini_dict_str)

    def eval_preconditioner_ini(self):
    
        sp_jac_trap_eval(self.sp_jac_ini.data,self.x,self.y_run,self.u_run,self.p,self.Dt)
    
        P_slu = spilu(self.sp_jac_ini,
                  fill_factor=self.fill_factor_ini,
                  drop_tol=self.drop_tol_ini,
                  drop_rule = self.drop_rule_ini)
    
        self.P_slu = P_slu
        P_d,P_i,P_p,perm_r,perm_c = slu2pydae(P_slu)   
        self.P_d = P_d
        self.P_i = P_i
        self.P_p = P_p
    
        self.perm_r = perm_r
        self.perm_c = perm_c
            
    
    def eval_preconditioner_run(self):
    
        sp_jac_trap_eval(self.J_run_d,self.x,self.y_run,self.u_run,self.p,self.Dt)
    
        self.sp_jac_trap.data = self.J_run_d 
        P_slu_run = spilu(self.sp_jac_trap,
                          fill_factor=self.fill_factor_run,
                          drop_tol=self.drop_tol_run,
                          drop_rule = self.drop_rule_run)
    
        self.P_slu_run = P_slu_run
        P_d,P_i,P_p,perm_r,perm_c = slu2pydae(P_slu_run)   
        self.P_run_d = P_d
        self.P_run_i = P_i
        self.P_run_p = P_p
    
        self.perm_run_r = perm_r
        self.perm_run_c = perm_c
        
    def sprun(self,t_end,up_dict):
        
        for item in up_dict:
            self.set_value(item,up_dict[item])
    
        t = self.t
        p = self.p
        it = self.it
        it_store = self.it_store
        xy = self.xy
        u = self.u_run
        self.iparams_run = np.zeros(10,dtype=np.float64)
    
        t,it,it_store,xy = spdaesolver(t,t_end,it,it_store,xy,u,p,
                                  self.jac_trap,
                                  self.J_run_d,self.J_run_i,self.J_run_p,
                                  self.P_run_d,self.P_run_i,self.P_run_p,self.perm_run_r,self.perm_run_c,
                                  self.Time,
                                  self.X,
                                  self.Y,
                                  self.Z,
                                  self.iters,
                                  self.Dt,
                                  self.N_x,
                                  self.N_y,
                                  self.N_z,
                                  self.decimation,
                                  self.iparams_run,
                                  max_it=self.max_it,itol=self.max_it,store=self.store,
                                  lmax_it=self.lmax_it,ltol=self.ltol,ldamp=self.ldamp,mode=self.mode)
    
        self.t = t
        self.it = it
        self.it_store = it_store
        self.xy = xy
            
    def spini(self,up_dict,xy_0={}):
    
        self.it = 0
        self.it_store = 0
        self.t = 0.0
    
        for item in up_dict:
            self.set_value(item,up_dict[item])
    
        if type(xy_0) == dict:
            xy_0_dict = xy_0
            self.dict2xy0(xy_0_dict)
    
        if type(xy_0) == str:
            if xy_0 == 'eval':
                N_x = self.N_x
                self.xy_0_new = np.copy(self.xy_0)*0
                xy0_eval(self.xy_0_new[:N_x],self.xy_0_new[N_x:],self.u_ini,self.p)
                self.xy_0_evaluated = np.copy(self.xy_0_new)
                self.xy_0 = np.copy(self.xy_0_new)
            else:
                self.load_xy_0(file_name = xy_0)
    
        self.xy_ini = self.spss_ini()
        self.ini2run()
        #jac_run_ss_eval_xy(self.jac_run,self.x,self.y_run,self.u_run,self.p)
        #jac_run_ss_eval_up(self.jac_run,self.x,self.y_run,self.u_run,self.p)

        
    def spss_ini(self):
        J_d,J_i,J_p = csr2pydae(self.sp_jac_ini)
        
        xy_ini,it,iparams = spsstate(self.xy,self.u_ini,self.p,
                 J_d,J_i,J_p,
                 self.P_d,self.P_i,self.P_p,self.perm_r,self.perm_c,
                 self.N_x,self.N_y,
                 max_it=self.max_it,tol=self.itol,
                 lmax_it=self.lmax_it_ini,
                 ltol=self.ltol_ini,
                 ldamp=self.ldamp)

 
        self.xy_ini = xy_ini
        self.N_iters = it
        self.iparams = iparams
    
        return xy_ini

    #def import_cffi(self):
        

        

           
            



def daesolver_sp(t,t_end,it,it_store,xy,u,p,sp_jac_trap,T,X,Y,Z,iters,Dt,N_x,N_y,N_z,decimation,max_it=50,itol=1e-8,store=1): 

    fg = np.zeros((N_x+N_y,1),dtype=np.float64)
    fg_i = np.zeros((N_x+N_y),dtype=np.float64)
    x = xy[:N_x]
    y = xy[N_x:]
    fg = np.zeros((N_x+N_y,),dtype=np.float64)
    f = fg[:N_x]
    g = fg[N_x:]
    h = np.zeros((N_z),dtype=np.float64)
    sp_jac_trap_eval_up(sp_jac_trap.data,x,y,u,p,Dt,xyup=1)
    
    if it == 0:
        f_run_eval(f,x,y,u,p)
        h_eval(h,x,y,u,p)
        it_store = 0  
        T[0] = t 
        X[0,:] = x  
        Y[0,:] = y  
        Z[0,:] = h  

    while t<t_end: 
        it += 1
        t += Dt

        f_run_eval(f,x,y,u,p)
        g_run_eval(g,x,y,u,p)

        x_0 = np.copy(x) 
        y_0 = np.copy(y) 
        f_0 = np.copy(f) 
        g_0 = np.copy(g) 
            
        for iti in range(max_it):
            f_run_eval(f,x,y,u,p)
            g_run_eval(g,x,y,u,p)
            sp_jac_trap_eval(sp_jac_trap.data,x,y,u,p,Dt,xyup=1)            

            f_n_i = x - x_0 - 0.5*Dt*(f+f_0) 

            fg_i[:N_x] = f_n_i
            fg_i[N_x:] = g
            
            Dxy_i = spsolve(sp_jac_trap,-fg_i) 

            x = x + Dxy_i[:N_x]
            y = y + Dxy_i[N_x:]              

            # iteration stop
            max_relative = 0.0
            for it_var in range(N_x+N_y):
                abs_value = np.abs(xy[it_var])
                if abs_value < 0.001:
                    abs_value = 0.001
                relative_error = np.abs(Dxy_i[it_var])/abs_value

                if relative_error > max_relative: max_relative = relative_error

            if max_relative<itol:
                break
                
        h_eval(h,x,y,u,p)
        xy[:N_x] = x
        xy[N_x:] = y
        
        # store in channels 
        if store == 1:
            if it >= it_store*decimation: 
                T[it_store+1] = t 
                X[it_store+1,:] = x 
                Y[it_store+1,:] = y
                Z[it_store+1,:] = h
                iters[it_store+1] = iti
                it_store += 1 

    return t,it,it_store,xy




@numba.njit()
def sprichardson(A_d,A_i,A_p,b,P_d,P_i,P_p,perm_r,perm_c,x,iparams,damp=1.0,max_it=100,tol=1e-3):
    N_A = A_p.shape[0]-1
    f = np.zeros(N_A)
    for it in range(max_it):
        spMvmul(N_A,A_d,A_i,A_p,x,f) 
        f -= b                          # A@x-b
        x = x - damp*splu_solve(P_d,P_i,P_p,perm_r,perm_c,f)   
        if np.linalg.norm(f,2) < tol: break
    iparams[0] = it
    return x
    
    

@numba.njit()
def spsstate(xy,u,p,
             J_d,J_i,J_p,
             P_d,P_i,P_p,perm_r,perm_c,
             N_x,N_y,
             max_it=50,tol=1e-8,
             lmax_it=20,ltol=1e-8,ldamp=1.0):
    
   
    x = xy[:N_x]
    y = xy[N_x:]
    fg = np.zeros((N_x+N_y,),dtype=np.float64)
    f = fg[:N_x]
    g = fg[N_x:]
    iparams = np.array([0],dtype=np.int64)    
    
    f_c_ptr=ffi.from_buffer(np.ascontiguousarray(f))
    g_c_ptr=ffi.from_buffer(np.ascontiguousarray(g))
    x_c_ptr=ffi.from_buffer(np.ascontiguousarray(x))
    y_c_ptr=ffi.from_buffer(np.ascontiguousarray(y))
    u_c_ptr=ffi.from_buffer(np.ascontiguousarray(u))
    p_c_ptr=ffi.from_buffer(np.ascontiguousarray(p))
    J_d_ptr=ffi.from_buffer(np.ascontiguousarray(J_d))

    sp_jac_ini_num_eval(J_d_ptr,x_c_ptr,y_c_ptr,u_c_ptr,p_c_ptr,1.0)
    sp_jac_ini_up_eval(J_d_ptr,x_c_ptr,y_c_ptr,u_c_ptr,p_c_ptr,1.0)
    
    #sp_jac_ini_eval_up(J_d,x,y,u,p,0.0)

    Dxy = np.zeros(N_x + N_y)
    for it in range(max_it):
        
        x = xy[:N_x]
        y = xy[N_x:]   
       
        sp_jac_ini_xy_eval(J_d_ptr,x_c_ptr,y_c_ptr,u_c_ptr,p_c_ptr,1.0)

        
        f_ini_eval(f_c_ptr,x_c_ptr,y_c_ptr,u_c_ptr,p_c_ptr,1.0)
        g_ini_eval(g_c_ptr,x_c_ptr,y_c_ptr,u_c_ptr,p_c_ptr,1.0)
        
        #f_ini_eval(f,x,y,u,p)
        #g_ini_eval(g,x,y,u,p)
        
        fg[:N_x] = f
        fg[N_x:] = g
               
        Dxy = sprichardson(J_d,J_i,J_p,-fg,P_d,P_i,P_p,perm_r,perm_c,Dxy,iparams,damp=ldamp,max_it=lmax_it,tol=ltol)
   
        xy += Dxy
        #if np.max(np.abs(fg))<tol: break
        if np.linalg.norm(fg,np.inf)<tol: break

    return xy,it,iparams


@numba.njit() 
def daesolver(t,t_end,it,it_store,xy,u,p,jac_trap,T,X,Y,Z,iters,Dt,N_x,N_y,N_z,decimation,max_it=50,itol=1e-8,store=1): 


    fg = np.zeros((N_x+N_y,1),dtype=np.float64)
    fg_i = np.zeros((N_x+N_y),dtype=np.float64)
    x = xy[:N_x]
    y = xy[N_x:]
    fg = np.zeros((N_x+N_y,),dtype=np.float64)
    f = fg[:N_x]
    g = fg[N_x:]
    h = np.zeros((N_z),dtype=np.float64)
    
    f_ptr=ffi.from_buffer(np.ascontiguousarray(f))
    g_ptr=ffi.from_buffer(np.ascontiguousarray(g))
    h_ptr=ffi.from_buffer(np.ascontiguousarray(h))
    x_ptr=ffi.from_buffer(np.ascontiguousarray(x))
    y_ptr=ffi.from_buffer(np.ascontiguousarray(y))
    u_ptr=ffi.from_buffer(np.ascontiguousarray(u))
    p_ptr=ffi.from_buffer(np.ascontiguousarray(p))

    jac_trap_ptr=ffi.from_buffer(np.ascontiguousarray(jac_trap))
    
    de_jac_trap_num_eval(jac_trap_ptr,x_ptr,y_ptr,u_ptr,p_ptr,Dt)    
    de_jac_trap_up_eval(jac_trap_ptr,x_ptr,y_ptr,u_ptr,p_ptr,Dt) 
    de_jac_trap_xy_eval(jac_trap_ptr,x_ptr,y_ptr,u_ptr,p_ptr,Dt) 
    
    if it == 0:
        f_run_eval(f_ptr,x_ptr,y_ptr,u_ptr,p_ptr,Dt)
        g_run_eval(g_ptr,x_ptr,y_ptr,u_ptr,p_ptr,Dt)
        h_eval(h_ptr,x_ptr,y_ptr,u_ptr,p_ptr,Dt)
        it_store = 0  
        T[0] = t 
        X[0,:] = x  
        Y[0,:] = y  
        Z[0,:] = h  

    while t<t_end: 
        it += 1
        t += Dt

        f_run_eval(f_ptr,x_ptr,y_ptr,u_ptr,p_ptr,Dt)
        g_run_eval(g_ptr,x_ptr,y_ptr,u_ptr,p_ptr,Dt)

        x_0 = np.copy(x) 
        y_0 = np.copy(y) 
        f_0 = np.copy(f) 
        g_0 = np.copy(g) 
            
        for iti in range(max_it):
            f_run_eval(f_ptr,x_ptr,y_ptr,u_ptr,p_ptr,Dt)
            g_run_eval(g_ptr,x_ptr,y_ptr,u_ptr,p_ptr,Dt)
            de_jac_trap_xy_eval(jac_trap_ptr,x_ptr,y_ptr,u_ptr,p_ptr,Dt) 

            f_n_i = x - x_0 - 0.5*Dt*(f+f_0) 

            fg_i[:N_x] = f_n_i
            fg_i[N_x:] = g
            
            Dxy_i = np.linalg.solve(-jac_trap,fg_i) 

            x += Dxy_i[:N_x]
            y += Dxy_i[N_x:] 
            
            #print(Dxy_i)

            # iteration stop
            max_relative = 0.0
            for it_var in range(N_x+N_y):
                abs_value = np.abs(xy[it_var])
                if abs_value < 0.001:
                    abs_value = 0.001
                relative_error = np.abs(Dxy_i[it_var])/abs_value

                if relative_error > max_relative: max_relative = relative_error

            if max_relative<itol:
                break
                
        h_eval(h_ptr,x_ptr,y_ptr,u_ptr,p_ptr,Dt)
        xy[:N_x] = x
        xy[N_x:] = y
        
        # store in channels 
        if store == 1:
            if it >= it_store*decimation: 
                T[it_store+1] = t 
                X[it_store+1,:] = x 
                Y[it_store+1,:] = y
                Z[it_store+1,:] = h
                iters[it_store+1] = iti
                it_store += 1 

    return t,it,it_store,xy
    
@numba.njit() 
def spdaesolver(t,t_end,it,it_store,xy,u,p,jac_trap,
                J_d,J_i,J_p,
                P_d,P_i,P_p,perm_r,perm_c,
                T,X,Y,Z,iters,Dt,N_x,N_y,N_z,decimation,
                iparams,
                max_it=50,itol=1e-8,store=1,
                lmax_it=20,ltol=1e-4,ldamp=1.0,mode=0):

    fg = np.zeros((N_x+N_y,1),dtype=np.float64)
    fg_i = np.zeros((N_x+N_y),dtype=np.float64)
    x = xy[:N_x]
    y = xy[N_x:]
    fg = np.zeros((N_x+N_y,),dtype=np.float64)
    f = fg[:N_x]
    g = fg[N_x:]
    h = np.zeros((N_z),dtype=np.float64)
    Dxy_i_0 = np.zeros(N_x+N_y,dtype=np.float64)
    
    f_ptr=ffi.from_buffer(np.ascontiguousarray(f))
    g_ptr=ffi.from_buffer(np.ascontiguousarray(g))
    h_ptr=ffi.from_buffer(np.ascontiguousarray(h))
    x_ptr=ffi.from_buffer(np.ascontiguousarray(x))
    y_ptr=ffi.from_buffer(np.ascontiguousarray(y))
    u_ptr=ffi.from_buffer(np.ascontiguousarray(u))
    p_ptr=ffi.from_buffer(np.ascontiguousarray(p))

    J_d_ptr=ffi.from_buffer(np.ascontiguousarray(J_d))
    
    sp_jac_trap_num_eval(J_d_ptr,x_ptr,y_ptr,u_ptr,p_ptr,Dt)    
    sp_jac_trap_up_eval(J_d_ptr,x_ptr,y_ptr,u_ptr,p_ptr,Dt) 
    sp_jac_trap_xy_eval(J_d_ptr,x_ptr,y_ptr,u_ptr,p_ptr,Dt) 
    
    if it == 0:
        f_run_eval(f_ptr,x_ptr,y_ptr,u_ptr,p_ptr,Dt)
        g_run_eval(g_ptr,x_ptr,y_ptr,u_ptr,p_ptr,Dt)
        h_eval(h_ptr,x_ptr,y_ptr,u_ptr,p_ptr,Dt)
        it_store = 0  
        T[0] = t 
        X[0,:] = x  
        Y[0,:] = y  
        Z[0,:] = h  

    while t<t_end: 
        it += 1
        t += Dt

        f_run_eval(f_ptr,x_ptr,y_ptr,u_ptr,p_ptr,Dt)
        g_run_eval(g_ptr,x_ptr,y_ptr,u_ptr,p_ptr,Dt)

        x_0 = np.copy(x) 
        y_0 = np.copy(y) 
        f_0 = np.copy(f) 
        g_0 = np.copy(g) 
            
        for iti in range(max_it):
            f_run_eval(f_ptr,x_ptr,y_ptr,u_ptr,p_ptr,Dt)
            g_run_eval(g_ptr,x_ptr,y_ptr,u_ptr,p_ptr,Dt)
            sp_jac_trap_xy_eval(J_d_ptr,x_ptr,y_ptr,u_ptr,p_ptr,Dt) 

            f_n_i = x - x_0 - 0.5*Dt*(f+f_0) 

            fg_i[:N_x] = f_n_i
            fg_i[N_x:] = g
            
            #Dxy_i = np.linalg.solve(-jac_trap,fg_i) 
            Dxy_i = sprichardson(J_d,J_i,J_p,-fg_i,P_d,P_i,P_p,perm_r,perm_c,
                                 Dxy_i_0,iparams,damp=ldamp,max_it=lmax_it,tol=ltol)

            x += Dxy_i[:N_x]
            y += Dxy_i[N_x:] 
            
            #print(Dxy_i)

            # iteration stop
            max_relative = 0.0
            for it_var in range(N_x+N_y):
                abs_value = np.abs(xy[it_var])
                if abs_value < 0.001:
                    abs_value = 0.001
                relative_error = np.abs(Dxy_i[it_var])/abs_value

                if relative_error > max_relative: max_relative = relative_error

            if max_relative<itol:
                break
                
        h_eval(h_ptr,x_ptr,y_ptr,u_ptr,p_ptr,Dt)
        xy[:N_x] = x
        xy[N_x:] = y
        
        # store in channels 
        if store == 1:
            if it >= it_store*decimation: 
                T[it_store+1] = t 
                X[it_store+1,:] = x 
                Y[it_store+1,:] = y
                Z[it_store+1,:] = h
                iters[it_store+1] = iti
                it_store += 1 

    return t,it,it_store,xy


@cuda.jit()
def ode_solve(x,u,p,f_run,u_idxs,z_i,z,sim):

    N_i,N_j,N_x,N_z,Dt = sim

    # index of thread on GPU:
    i = cuda.grid(1)

    if i < x.size:
        for j in range(N_j):
            f_run_eval(f_run[i,:],x[i,:],u[i,u_idxs[j],:],p[i,:])
            for k in range(N_x):
              x[i,k] +=  Dt*f_run[i,k]

            # outputs in time range
            #z[i,j] = u[i,idxs[j],0]
            z[i,j] = x[i,1]
        h_eval(z_i[i,:],x[i,:],u[i,u_idxs[j],:],p[i,:])
        
def csr2pydae(A_csr):
    '''
    From scipy CSR to the three vectors:
    
    - data
    - indices
    - indptr
    
    '''
    
    A_d = A_csr.data
    A_i = A_csr.indices
    A_p = A_csr.indptr
    
    return A_d,A_i,A_p
    
def slu2pydae(P_slu):
    '''
    From SupderLU matrix to the three vectors:
    
    - data
    - indices
    - indptr
    
    and the premutation vectors:
    
    - perm_r
    - perm_c
    
    '''
    N = P_slu.shape[0]
    P_slu_full = P_slu.L.A - sspa.eye(N,format='csr') + P_slu.U.A
    perm_r = P_slu.perm_r
    perm_c = P_slu.perm_c
    P_csr = sspa.csr_matrix(P_slu_full)
    
    P_d = P_csr.data
    P_i = P_csr.indices
    P_p = P_csr.indptr
    
    return P_d,P_i,P_p,perm_r,perm_c

@numba.njit(cache=True)
def spMvmul(N,A_data,A_indices,A_indptr,x,y):
    '''
    y = A @ x
    
    with A in sparse CRS form
    '''
    #y = np.zeros(x.shape[0])
    for i in range(N):
        y[i] = 0.0
        for j in range(A_indptr[i],A_indptr[i + 1]):
            y[i] = y[i] + A_data[j]*x[A_indices[j]]
            
            
@numba.njit(cache=True)
def splu_solve(LU_d,LU_i,LU_p,perm_r,perm_c,b):
    N = len(b)
    y = np.zeros(N)
    x = np.zeros(N)
    z = np.zeros(N)
    bp = np.zeros(N)
    
    for i in range(N): 
        bp[perm_r[i]] = b[i]
        
    for i in range(N): 
        y[i] = bp[i]
        for j in range(LU_p[i],LU_p[i+1]):
            if LU_i[j]>i-1: break
            y[i] -= LU_d[j] * y[LU_i[j]]

    for i in range(N-1,-1,-1): #(int i = N - 1; i >= 0; i--) 
        z[i] = y[i]
        den = 0.0
        for j in range(LU_p[i],LU_p[i+1]): #(int k = i + 1; k < N; k++)
            if LU_i[j] > i:
                z[i] -= LU_d[j] * z[LU_i[j]]
            if LU_i[j] == i: den = LU_d[j]
        z[i] = z[i]/den
 
    for i in range(N):
        x[i] = z[perm_c[i]]
        
    return x



@numba.njit("float64[:,:](float64[:,:],float64[:],float64[:],float64[:],float64[:],float64)")
def de_jac_ini_eval(de_jac_ini,x,y,u,p,Dt):   
    '''
    Computes the dense full initialization jacobian:
    
    jac_ini = [[Fx_ini, Fy_ini],
               [Gx_ini, Gy_ini]]
                
    for the given x,y,u,p vectors and Dt time increment.
    
    Parameters
    ----------
    de_jac_ini : (N, N) array_like
                  Input data.
    x : (N_x,) array_like
        Vector with dynamical states.
    y : (N_y,) array_like
        Vector with algebraic states (ini problem).
    u : (N_u,) array_like
        Vector with inputs (ini problem). 
    p : (N_p,) array_like
        Vector with parameters. 
        
    with N = N_x+N_y
 
    Returns
    -------
    
    de_jac_ini : (N, N) array_like
                  Updated matrix.    
    
    '''
    
    de_jac_ini_ptr=ffi.from_buffer(np.ascontiguousarray(de_jac_ini))
    x_c_ptr=ffi.from_buffer(np.ascontiguousarray(x))
    y_c_ptr=ffi.from_buffer(np.ascontiguousarray(y))
    u_c_ptr=ffi.from_buffer(np.ascontiguousarray(u))
    p_c_ptr=ffi.from_buffer(np.ascontiguousarray(p))

    de_jac_ini_num_eval(de_jac_ini_ptr,x_c_ptr,y_c_ptr,u_c_ptr,p_c_ptr,Dt)
    de_jac_ini_up_eval( de_jac_ini_ptr,x_c_ptr,y_c_ptr,u_c_ptr,p_c_ptr,Dt)
    de_jac_ini_xy_eval( de_jac_ini_ptr,x_c_ptr,y_c_ptr,u_c_ptr,p_c_ptr,Dt)
    
    return de_jac_ini

@numba.njit("float64[:,:](float64[:,:],float64[:],float64[:],float64[:],float64[:],float64)")
def de_jac_trap_eval(de_jac_trap,x,y,u,p,Dt):   
    '''
    Computes the dense full trapezoidal jacobian:
    
    jac_trap = [[eye - 0.5*Dt*Fx_run, -0.5*Dt*Fy_run],
                [             Gx_run,         Gy_run]]
                
    for the given x,y,u,p vectors and Dt time increment.
    
    Parameters
    ----------
    de_jac_trap : (N, N) array_like
                  Input data.
    x : (N_x,) array_like
        Vector with dynamical states.
    y : (N_y,) array_like
        Vector with algebraic states (run problem).
    u : (N_u,) array_like
        Vector with inputs (run problem). 
    p : (N_p,) array_like
        Vector with parameters. 
 
    Returns
    -------
    
    de_jac_trap : (N, N) array_like
                  Updated matrix.    
    
    '''
        
    de_jac_trap_ptr = ffi.from_buffer(np.ascontiguousarray(de_jac_trap))
    x_c_ptr = ffi.from_buffer(np.ascontiguousarray(x))
    y_c_ptr = ffi.from_buffer(np.ascontiguousarray(y))
    u_c_ptr = ffi.from_buffer(np.ascontiguousarray(u))
    p_c_ptr = ffi.from_buffer(np.ascontiguousarray(p))

    de_jac_trap_num_eval(de_jac_trap_ptr,x_c_ptr,y_c_ptr,u_c_ptr,p_c_ptr,Dt)
    de_jac_trap_up_eval( de_jac_trap_ptr,x_c_ptr,y_c_ptr,u_c_ptr,p_c_ptr,Dt)
    de_jac_trap_xy_eval( de_jac_trap_ptr,x_c_ptr,y_c_ptr,u_c_ptr,p_c_ptr,Dt)
    
    return de_jac_trap


@numba.njit("float64[:](float64[:],float64[:],float64[:],float64[:],float64[:],float64)")
def sp_jac_trap_eval(sp_jac_trap,x,y,u,p,Dt):   
    '''
    Computes the sparse full trapezoidal jacobian:
    
    jac_trap = [[eye - 0.5*Dt*Fx_run, -0.5*Dt*Fy_run],
                [             Gx_run,         Gy_run]]
                
    for the given x,y,u,p vectors and Dt time increment.
    
    Parameters
    ----------
    sp_jac_trap : (Nnz,) array_like
                  Input data.
    x : (N_x,) array_like
        Vector with dynamical states.
    y : (N_y,) array_like
        Vector with algebraic states (run problem).
    u : (N_u,) array_like
        Vector with inputs (run problem). 
    p : (N_p,) array_like
        Vector with parameters. 
        
    with Nnz the number of non-zeros elements in the jacobian.
 
    Returns
    -------
    
    sp_jac_trap : (Nnz,) array_like
                  Updated matrix.    
    
    '''        
    sp_jac_trap_ptr=ffi.from_buffer(np.ascontiguousarray(sp_jac_trap))
    x_c_ptr=ffi.from_buffer(np.ascontiguousarray(x))
    y_c_ptr=ffi.from_buffer(np.ascontiguousarray(y))
    u_c_ptr=ffi.from_buffer(np.ascontiguousarray(u))
    p_c_ptr=ffi.from_buffer(np.ascontiguousarray(p))

    sp_jac_trap_num_eval(sp_jac_trap_ptr,x_c_ptr,y_c_ptr,u_c_ptr,p_c_ptr,Dt)
    sp_jac_trap_up_eval( sp_jac_trap_ptr,x_c_ptr,y_c_ptr,u_c_ptr,p_c_ptr,Dt)
    sp_jac_trap_xy_eval( sp_jac_trap_ptr,x_c_ptr,y_c_ptr,u_c_ptr,p_c_ptr,Dt)
    
    return sp_jac_trap

@numba.njit("float64[:](float64[:],float64[:],float64[:],float64[:],float64[:],float64)")
def sp_jac_ini_eval(sp_jac_ini,x,y,u,p,Dt):   
    '''
    Computes the SPARSE full initialization jacobian:
    
    jac_ini = [[Fx_ini, Fy_ini],
               [Gx_ini, Gy_ini]]
                
    for the given x,y,u,p vectors and Dt time increment.
    
    Parameters
    ----------
    de_jac_ini : (N, N) array_like
                  Input data.
    x : (N_x,) array_like
        Vector with dynamical states.
    y : (N_y,) array_like
        Vector with algebraic states (ini problem).
    u : (N_u,) array_like
        Vector with inputs (ini problem). 
    p : (N_p,) array_like
        Vector with parameters. 
        
    with N = N_x+N_y
 
    Returns
    -------
    
    de_jac_ini : (N, N) array_like
                  Updated matrix.    
    
    '''
    
    sp_jac_ini_ptr=ffi.from_buffer(np.ascontiguousarray(sp_jac_ini))
    x_c_ptr=ffi.from_buffer(np.ascontiguousarray(x))
    y_c_ptr=ffi.from_buffer(np.ascontiguousarray(y))
    u_c_ptr=ffi.from_buffer(np.ascontiguousarray(u))
    p_c_ptr=ffi.from_buffer(np.ascontiguousarray(p))

    sp_jac_ini_num_eval(sp_jac_ini_ptr,x_c_ptr,y_c_ptr,u_c_ptr,p_c_ptr,Dt)
    sp_jac_ini_up_eval( sp_jac_ini_ptr,x_c_ptr,y_c_ptr,u_c_ptr,p_c_ptr,Dt)
    sp_jac_ini_xy_eval( sp_jac_ini_ptr,x_c_ptr,y_c_ptr,u_c_ptr,p_c_ptr,Dt)
    
    return sp_jac_ini


@numba.njit()
def sstate(xy,u,p,jac_ini_ss,N_x,N_y,max_it=50,tol=1e-8):
    
    x = xy[:N_x]
    y = xy[N_x:]
    fg = np.zeros((N_x+N_y,),dtype=np.float64)
    f = fg[:N_x]
    g = fg[N_x:]

    f_c_ptr=ffi.from_buffer(np.ascontiguousarray(f))
    g_c_ptr=ffi.from_buffer(np.ascontiguousarray(g))
    x_c_ptr=ffi.from_buffer(np.ascontiguousarray(x))
    y_c_ptr=ffi.from_buffer(np.ascontiguousarray(y))
    u_c_ptr=ffi.from_buffer(np.ascontiguousarray(u))
    p_c_ptr=ffi.from_buffer(np.ascontiguousarray(p))
    jac_ini_ss_ptr=ffi.from_buffer(np.ascontiguousarray(jac_ini_ss))

    de_jac_ini_num_eval(jac_ini_ss_ptr,x_c_ptr,y_c_ptr,u_c_ptr,p_c_ptr,1.0)
    de_jac_ini_up_eval(jac_ini_ss_ptr,x_c_ptr,y_c_ptr,u_c_ptr,p_c_ptr,1.0)

    for it in range(max_it):
        de_jac_ini_xy_eval(jac_ini_ss_ptr,x_c_ptr,y_c_ptr,u_c_ptr,p_c_ptr,1.0)
        f_ini_eval(f_c_ptr,x_c_ptr,y_c_ptr,u_c_ptr,p_c_ptr,1.0)
        g_ini_eval(g_c_ptr,x_c_ptr,y_c_ptr,u_c_ptr,p_c_ptr,1.0)
        fg[:N_x] = f
        fg[N_x:] = g
        xy += np.linalg.solve(jac_ini_ss,-fg)
        if np.max(np.abs(fg))<tol: break

    return xy,it


@numba.njit("float64[:](float64[:],float64[:],float64[:],float64[:],float64[:],float64)")
def c_h_eval(z,x,y,u,p,Dt):   
    '''
    Computes the SPARSE full initialization jacobian:
    
    jac_ini = [[Fx_ini, Fy_ini],
               [Gx_ini, Gy_ini]]
                
    for the given x,y,u,p vectors and Dt time increment.
    
    Parameters
    ----------
    de_jac_ini : (N, N) array_like
                  Input data.
    x : (N_x,) array_like
        Vector with dynamical states.
    y : (N_y,) array_like
        Vector with algebraic states (ini problem).
    u : (N_u,) array_like
        Vector with inputs (ini problem). 
    p : (N_p,) array_like
        Vector with parameters. 
        
    with N = N_x+N_y
 
    Returns
    -------
    
    de_jac_ini : (N, N) array_like
                  Updated matrix.    
    
    '''
    
    z_c_ptr=ffi.from_buffer(np.ascontiguousarray(z))
    x_c_ptr=ffi.from_buffer(np.ascontiguousarray(x))
    y_c_ptr=ffi.from_buffer(np.ascontiguousarray(y))
    u_c_ptr=ffi.from_buffer(np.ascontiguousarray(u))
    p_c_ptr=ffi.from_buffer(np.ascontiguousarray(p))

    h_eval(z_c_ptr,x_c_ptr,y_c_ptr,u_c_ptr,p_c_ptr,Dt)
    
    return z

def sp_jac_ini_vectors():

    sp_jac_ini_ia = [0, 1, 0, 1, 5, 8, 9, 22, 23, 2, 22, 26, 3, 23, 4, 8, 5, 27, 1, 7, 8, 9, 10, 11, 24, 8, 9, 10, 11, 25, 8, 9, 10, 11, 12, 13, 8, 9, 10, 11, 12, 13, 10, 11, 12, 13, 14, 15, 18, 19, 10, 11, 12, 13, 14, 15, 18, 19, 12, 13, 14, 15, 16, 17, 12, 13, 14, 15, 16, 17, 14, 15, 16, 17, 14, 15, 16, 17, 12, 13, 18, 19, 20, 21, 12, 13, 18, 19, 20, 21, 18, 19, 20, 21, 18, 19, 20, 21, 0, 2, 8, 9, 22, 23, 0, 3, 8, 9, 22, 23, 0, 8, 9, 22, 23, 24, 0, 8, 9, 22, 23, 25, 4, 26, 1, 6, 27]
    sp_jac_ini_ja = [0, 2, 9, 12, 14, 16, 18, 19, 20, 25, 30, 36, 42, 50, 58, 64, 70, 74, 78, 84, 90, 94, 98, 104, 110, 116, 122, 124, 127]
    sp_jac_ini_nia = 28
    sp_jac_ini_nja = 28
    return sp_jac_ini_ia, sp_jac_ini_ja, sp_jac_ini_nia, sp_jac_ini_nja 

def sp_jac_trap_vectors():

    sp_jac_trap_ia = [0, 1, 0, 1, 5, 8, 9, 22, 23, 2, 22, 26, 3, 23, 4, 8, 5, 27, 1, 6, 7, 8, 9, 10, 11, 24, 8, 9, 10, 11, 25, 8, 9, 10, 11, 12, 13, 8, 9, 10, 11, 12, 13, 10, 11, 12, 13, 14, 15, 18, 19, 10, 11, 12, 13, 14, 15, 18, 19, 12, 13, 14, 15, 16, 17, 12, 13, 14, 15, 16, 17, 14, 15, 16, 17, 14, 15, 16, 17, 12, 13, 18, 19, 20, 21, 12, 13, 18, 19, 20, 21, 18, 19, 20, 21, 18, 19, 20, 21, 0, 2, 8, 9, 22, 23, 0, 3, 8, 9, 22, 23, 0, 8, 9, 22, 23, 24, 0, 8, 9, 22, 23, 25, 4, 26, 1, 6, 27]
    sp_jac_trap_ja = [0, 2, 9, 12, 14, 16, 18, 20, 21, 26, 31, 37, 43, 51, 59, 65, 71, 75, 79, 85, 91, 95, 99, 105, 111, 117, 123, 125, 128]
    sp_jac_trap_nia = 28
    sp_jac_trap_nja = 28
    return sp_jac_trap_ia, sp_jac_trap_ja, sp_jac_trap_nia, sp_jac_trap_nja 



# -*- coding: utf-8 -*-
"""
Created on Thu August 10 23:52:55 2022

@author: jmmauricio
"""

import numpy as np
import sympy as sym
from pydae.urisi.vscs.ac3ph4wgfpi2 import ac3ph4wgfpi2
from pydae.urisi.vscs.ac3ph4wgfpi3 import ac3ph4wgfpi3
from pydae.urisi.vscs.ac_3ph_4w_l import ac_3ph_4w_l
from pydae.urisi.vscs.acdc_3ph_4w_vdc_q import acdc_3ph_4w_vdc_q
from pydae.urisi.vscs.acdc_3ph_4w_pq import acdc_3ph_4w_pq

def add_vscs(grid):

    buses = grid.data['buses']
    buses_list = [bus['name'] for bus in buses]

    for item in grid.data['vscs']:
        if item['type'] == 'ac_3ph_4w_l':
            ac_3ph_4w_l(grid,item)
        if item['type'] == 'ac3ph4wgfpi2':
            ac3ph4wgfpi2(grid,item)
        if item['type'] == 'ac3ph4wgfpi3':
            ac3ph4wgfpi3(grid,item)
        if item['type'] == 'acdc_3ph_4w_vdc_q': 
            acdc_3ph_4w_vdc_q(grid,item)
        if item['type'] == 'acdc_3ph_4w_pq': 
            acdc_3ph_4w_pq(grid,item)


        


from pydae.build_v2 import builder
import numpy as np
import sympy as sym
from sympy.matrices.sparsetools import _doktocsr
from sympy import SparseMatrix

def smib_dae():
    r"""
    Synchronous machine model of order 4 from Federico Milano book.

    """

    sin = sym.sin
    cos = sym.cos  

    # inputs
    V = sym.Symbol(f"V", real=True)
    theta = sym.Symbol(f"theta", real=True)
    p_m = sym.Symbol(f"p_m", real=True)
    v_f = sym.Symbol(f"v_f", real=True)  
        
    # dynamic states
    delta = sym.Symbol(f"delta", real=True)
    omega = sym.Symbol(f"omega", real=True)
    e1q = sym.Symbol(f"e1q", real=True)
    e1d = sym.Symbol(f"e1d", real=True)

    # algebraic states
    i_d = sym.Symbol(f"i_d", real=True)
    i_q = sym.Symbol(f"i_q", real=True)            
    p_g = sym.Symbol(f"p_g", real=True)
    q_g = sym.Symbol(f"q_g", real=True)

    # parameters
    S_n = sym.Symbol(f"S_n", real=True)
    Omega_b = sym.Symbol(f"Omega_b", real=True)            
    H = sym.Symbol(f"H", real=True)
    T1d0 = sym.Symbol(f"T1d0", real=True)
    T1q0 = sym.Symbol(f"T1q0", real=True)
    X_d = sym.Symbol(f"X_d", real=True)
    X_q = sym.Symbol(f"X_q", real=True)
    X1d = sym.Symbol(f"X1d", real=True)
    X1q = sym.Symbol(f"X1q", real=True)
    D = sym.Symbol(f"D", real=True)
    R_a = sym.Symbol(f"R_a", real=True)

    params_dict = {"S_n":100e6,
                   "X_d":1.8,"X1d":0.3, "T1d0":8.0,    
                   "X_q":1.7,"X1q":0.55,"T1q0":0.4,  
                   "R_a":0.01,"X_l": 0.2, 
                   "H":5.0,"D":1.0,
                   "Omega_b":2*np.pi*50}
    
    # auxiliar
    v_d = V*sin(delta - theta) 
    v_q = V*cos(delta - theta) 
    p_e = i_d*(v_d + R_a*i_d) + i_q*(v_q + R_a*i_q)     

    aux_dict = {"v_d":V*sin(delta - theta),
                "v_q":V*cos(delta - theta),
                "p_e":i_d*(v_d + R_a*i_d) + i_q*(v_q + R_a*i_q)}

    # dynamic equations            
    ddelta = Omega_b*(omega - 1.0) 
    domega = 1/(2*H)*(p_m - p_e - D*(omega - 1.0))
    de1q = 1/T1d0*(-e1q - (X_d - X1d)*i_d + v_f)
    de1d = 1/T1q0*(-e1d + (X_q - X1q)*i_q)

    # algebraic equations   
    g_i_d  = v_q + R_a*i_q + X1d*i_d - e1q
    g_i_q  = v_d + R_a*i_d - X1q*i_q - e1d
    g_p_g  = i_d*v_d + i_q*v_q - p_g  
    g_q_g  = i_d*v_q - i_q*v_d - q_g 
    
    # dae 
    f_list = [ddelta,domega,de1q,de1d]
    x_list = [ delta, omega, e1q, e1d]
    g_list = [g_i_d,g_i_q,g_p_g,g_q_g]
    y_ini_list = [  i_d,  i_q,  p_m,  q_g]
    y_run_list = [  i_d,  i_q,  p_g,  q_g]
   
    u_ini_dict = {}
    u_ini_dict.update({'v_f':1.5})
    u_ini_dict.update({'p_g':0.5})
    u_ini_dict.update({'V':1.0})
    u_ini_dict.update({'theta':0.0})

    u_run_dict = {}
    u_run_dict.update({'v_f':1.5})
    u_run_dict.update({'p_m':0.5})
    u_run_dict.update({'V':1.0})
    u_run_dict.update({'theta':0.0})

    xy_0_dict = {}
    xy_0_dict.update({'omega':1.0})
    xy_0_dict.update({'e1q':1.0})
    xy_0_dict.update({'i_q':0.5})

    # outputs
    h_dict = {}
    h_dict.update({f"p_e":p_e})
    h_dict.update({f"v_f":v_f})
    h_dict.update({f"p_m":p_m})
    
    sys_dict = {'name':'dae1','uz_jacs':True,
                'params_dict':params_dict,
                'f_list':f_list,
                'g_list':g_list,
                'x_list':x_list,
                'y_ini_list':y_ini_list,
                'y_run_list':y_run_list,
                'u_ini_dict':u_ini_dict,
                'u_run_dict':u_run_dict,
                'h_dict':h_dict,
                'aux_dict':aux_dict}

    return sys_dict

def smib_build():

    sys_dict = smib_dae()
    b = builder(sys_dict)
    b.dict2system()
    b.functions()
    b.jacobians()
    b.cwrite()
    b.compile()
    b.template()

def expected_builder():

    sys_dict = smib_dae()

    f = sym.Matrix(sys_dict['f_list'])
    g = sym.Matrix(sys_dict['g_list'])
    h = sym.Matrix([item for item in sys_dict['h_dict']])

    x = sym.Matrix(sys_dict['x_list'])
    y_ini = sym.Matrix(sys_dict['y_ini_list'])
    u_ini = sym.Matrix([item for item in sys_dict['u_ini_dict']])
    y_run = sym.Matrix(sys_dict['y_run_list'])
    u_run = sym.Matrix([item for item in sys_dict['u_run_dict']])

    F_x = f.jacobian(x)
    F_y_ini = f.jacobian(y_ini)
    F_u_ini = f.jacobian(u_ini)
    F_y_run = f.jacobian(y_run)
    F_u_run = f.jacobian(u_run)

    G_x = g.jacobian(x)
    G_y_ini = g.jacobian(y_ini)
    G_u_ini = g.jacobian(u_ini)
    G_y_run = g.jacobian(y_run)
    G_u_run = g.jacobian(u_run)

    H_x = h.jacobian(x)
    H_y = h.jacobian(y_run)
    H_u = h.jacobian(u_run)

    jac_ini = sym.Matrix([[F_x,F_y_ini],[G_x,G_y_ini]])
    jac_run = sym.Matrix([[F_x,F_y_run],[G_x,G_y_run]])

    N_x = F_x.shape[0]
    eye = sym.eye(N_x, real=True)
    Dt = sym.Symbol('Dt',real=True)
    jac_trap = sym.Matrix([[eye - 0.5*Dt*F_x, -0.5*Dt*F_y_run],[G_x,G_y_run]])

    tab = 3*' '
    string = ''
    string += 'import numpy as np\n\n'   
    string += 'def eval_dae(model):\n\n'
    string += f'{tab} sin = np.sin\n'  
    string += f'{tab} cos = np.cos\n'  
    string += f'{tab} Dt = 0.001\n'  

    for item in sys_dict['x_list']:
        string += f'{tab} {item} = model.get_value("{item}")\n'
    string += f'{tab} x = np.array({str(list(x))})\n\n'
    string += f"\n"

    for item in sys_dict['y_ini_list']:
        string += f'{tab} {item} = model.get_value("{item}")\n'
    string += f'{tab} y = np.array({str(list(y_ini))})\n\n'
    string += f"\n"

    for item in sys_dict['u_ini_dict']:
        string += f"{tab} {item} = {sys_dict['u_ini_dict'][item]}\n"
    string += f"\n"

    for item in sys_dict['params_dict']:
        string += f"{tab} {item} = {sys_dict['params_dict'][item]}\n"
    string += f"\n"

    for item in sys_dict['aux_dict']:
        string += f"{tab} {item} = {sys_dict['aux_dict'][item]}\n"
    string += f"\n"


    string += f"{tab} f = {str(f).replace('Matrix','np.array')}\n\n"
    string += f"{tab} g = {str(g).replace('Matrix','np.array')}\n\n"
    string += f"{tab} h = {str(h).replace('Matrix','np.array')}\n\n"

    string += f"{tab} F_x = {str(F_x).replace('Matrix','np.array')}\n\n"
    string += f"{tab} F_y_ini = {str(F_y_ini).replace('Matrix','np.array')}\n\n"
    string += f"{tab} F_u_ini = {str(F_u_ini).replace('Matrix','np.array')}\n\n"
    string += f"{tab} G_x = {str(G_x).replace('Matrix','np.array')}\n\n"
    string += f"{tab} G_y_ini = {str(G_y_ini).replace('Matrix','np.array')}\n\n"
    string += f"{tab} G_u_ini = {str(G_u_ini).replace('Matrix','np.array')}\n\n"
    string += f"{tab} H_x = {str(H_x).replace('Matrix','np.array')}\n\n"
    string += f"{tab} H_y = {str(H_y).replace('Matrix','np.array')}\n\n"
    string += f"{tab} H_u = {str(H_u).replace('Matrix','np.array')}\n\n"
    string += f"{tab} jac_ini = {str(jac_ini).replace('Matrix','np.array')}\n\n"
    string += f"{tab} jac_run = {str(jac_run).replace('Matrix','np.array')}\n\n"
    string += f"{tab} jac_trap = {str(jac_trap).replace('Matrix','np.array')}\n\n"

    string += "    expected_dict = {'f':f,'g':g,'h':h}\n"
    string += "    expected_dict.update({'F_x':F_x,'F_y_ini':F_y_ini,'F_u_ini':F_u_ini,'G_x':G_x})\n"
    string += "    expected_dict.update({'G_y':G_y_ini,'G_u_ini':G_u_ini,'H_x':H_x,'H_y':H_y,'H_u':H_u})\n"
    string += "    expected_dict.update({'jac_ini':jac_ini,'jac_run':jac_run,'jac_trap':jac_trap})\n\n"

    string += "    return expected_dict\n"

    with open('smib_expected.py','w') as fobj:
        fobj.write(string)

def test_jacobians():

    expected_builder()
    tol = 1e-12

    from smib_expected import eval_dae
    import dae1
    model = dae1.model()
    model.ini({},1)

    expected_dict =eval_dae(model) 

    # jac_ini
    from dae1 import de_jac_ini_eval,sp_jac_ini_eval
    N_x = model.N_x
    N_y = model.N_y
    N_xy = N_x+N_y
    Dt = 0.001

    ## de_jac_ini
    de_jac_ini = np.zeros((N_xy,N_xy))
    de_jac_ini_eval(de_jac_ini,model.x,model.y_ini,model.u_ini,model.p,Dt)

    assert np.max(np.abs(de_jac_ini-expected_dict['jac_ini'])) < tol

    ## sp_jac_ini
    from dae1 import sp_jac_ini_vectors
    sp_jac_ini_ia, sp_jac_ini_ja, sp_jac_ini_nia, sp_jac_ini_nja = sp_jac_ini_vectors()

    from scipy.sparse import csr_array
    jac_ini_data = np.zeros(len(sp_jac_ini_ia))
    sp_jac_ini = csr_array((np.zeros(len(jac_ini_data)),sp_jac_ini_ia,sp_jac_ini_ja),shape=(sp_jac_ini_nia,sp_jac_ini_nja))
    sp_jac_ini_eval(sp_jac_ini.data,model.x,model.y_ini,model.u_ini,model.p,Dt)

    assert np.max(np.abs(sp_jac_ini-expected_dict['jac_ini'])) < tol

    # jac_run
    from dae1 import de_jac_run_eval,sp_jac_run_eval
    N_x = model.N_x
    N_y = model.N_y
    N_xy = N_x+N_y
    Dt = 0.001

    ## de_jac_run
    de_jac_run = np.zeros((N_xy,N_xy))
    de_jac_run_eval(de_jac_run,model.x,model.y_run,model.u_run,model.p,Dt)

    assert np.max(np.abs(de_jac_run-expected_dict['jac_run'])) < tol

    ## sp_jac_run
    from dae1 import sp_jac_run_vectors
    sp_jac_run_ia, sp_jac_run_ja, sp_jac_run_nia, sp_jac_run_nja = sp_jac_run_vectors()

    from scipy.sparse import csr_array
    jac_run_data = np.zeros(len(sp_jac_run_ia))
    sp_jac_run = csr_array((np.zeros(len(jac_run_data)),sp_jac_run_ia,sp_jac_run_ja),shape=(sp_jac_run_nia,sp_jac_run_nja))
    sp_jac_run_eval(sp_jac_run.data,model.x,model.y_run,model.u_run,model.p,Dt)

    assert np.max(np.abs(sp_jac_run-expected_dict['jac_run'])) < tol


    # jac_trap
    from dae1 import de_jac_trap_eval,sp_jac_trap_eval
    N_x = model.N_x
    N_y = model.N_y
    N_xy = N_x+N_y
    Dt = 0.001

    ## de_jac_trap
    de_jac_trap = np.zeros((N_xy,N_xy))

    de_jac_trap_eval(de_jac_trap,model.x,model.y_run,model.u_run,model.p,Dt)
    assert np.max(np.abs(de_jac_trap-expected_dict['jac_trap'])) < tol

    ## sp_jac_trap
    from dae1 import sp_jac_trap_vectors
    sp_jac_trap_ia, sp_jac_trap_ja, sp_jac_trap_nia, sp_jac_trap_nja = sp_jac_trap_vectors()

    from scipy.sparse import csr_array
    jac_trap_data = np.zeros(len(sp_jac_trap_ia))
    sp_jac_trap = csr_array((np.zeros(len(jac_trap_data)),sp_jac_trap_ia,sp_jac_trap_ja),shape=(sp_jac_trap_nia,sp_jac_trap_nja))
    sp_jac_trap_eval(sp_jac_trap.data,model.x,model.y_run,model.u_run,model.p,Dt)
    assert np.max(np.abs(sp_jac_trap-expected_dict['jac_trap'])) < tol

smib_build()
test_jacobians()

                # 'f_list':f_list,
                # 'g_list':g_list,
                # 'x_list':x_list,
                # 'y_ini_list':y_list,
                # 'y_run_list':y_list,
                # 'u_run_dict':u_dict,
                # 'u_ini_dict':u_dict,
                # 'h_dict':h_dict
# y = sym.Matrix([[y_1],   
#                 [y_2],   
#                 [y_3],   
#                 [y_4]])

# u = sym.Matrix([[u_1],   
#                 [u_2]])


# f = sym.Matrix([[-p_f1*x_1 + p_f4*y_1 + p_u1*u_1 + x_3*y_1], 
#                 [-p_f2*x_3], 
#                 [-p_f3*x_2 + u_2*y_1]])
                
# g = sym.Matrix([[p_g1*y_2 + p_u2*u_1], 
#                 [p_g2*y_1 + x_2*y_1], 
#                 [p_g3*y_3 + x_3*y_2], 
#                 [p_u3*u_2 + u_1*y_4 + x_2*y_2]])

# h = sym.Matrix([[p_h1*x_1 + p_h2*y_1 + p_h3*y_4 + p_h4*u_1], 
#                 [u_2*x_1 + x_1*y_2 + x_2*y_1]])

# z = h

# F_x = f.jacobian(x)
# F_y = f.jacobian(y)
# F_u = f.jacobian(u)

# G_x = g.jacobian(x)
# G_y = g.jacobian(y)
# G_u = g.jacobian(u)

# H_x = h.jacobian(x)
# H_y = h.jacobian(y)
# H_u = h.jacobian(u)

# jac_ini_sym = sym.Matrix([[F_x,F_y],[G_x,G_y]]) 
# #print(jac_ini_sym)
# jac_ini_data,jac_ini_indices,jac_ini_indptr,jac_ini_shape = _doktocsr(SparseMatrix(jac_ini_sym))

# jac_ini_sym = sym.Matrix([[F_x,F_y],[G_x,G_y]]) 
# #print(jac_ini_sym)
# jac_ini_data,jac_ini_indices,jac_ini_indptr,jac_ini_shape = _doktocsr(SparseMatrix(jac_ini_sym))

# N_x = F_x.shape[0]
# eye = sym.eye(N_x, real=True)
# Dt = sym.Symbol('Dt',real=True)
# jac_trap = sym.Matrix([[eye - 0.5*Dt*F_x, -0.5*Dt*F_y],[G_x,G_y]])    
# print('jac_trap = ',jac_trap)
# jac_trap_data,jac_trap_indices,jac_trap_indptr,jac_trap_shape = _doktocsr(SparseMatrix(jac_trap_sym))

# print('F_x = ',F_x)
# print('F_y = ',F_y)
# print('F_u = ',F_u)

# print('G_x = ',G_x)
# print('G_y = ',G_y)
# print('G_u = ',G_u)

# print('H_x = ',H_x)
# print('H_y = ',H_y)
# print('H_u = ',H_u)


# p_f1,p_f2,p_f3,p_f4= 1.0,2.0,3.0,4.0
# p_g1,p_g2,p_g3,p_g4= 1.0,2.0,3.0,4.0
# p_u1,p_u2,p_u3,p_u4= 1.0,2.0,3.0,4.0
# p_h1,p_h2,p_h3,p_h4= 1.0,2.0,3.0,4.0

# params_dict = {
# 'p_f1':p_f1,'p_f2':p_f2,'p_f3':p_f3,'p_f4':p_f4,
# 'p_g1':p_g1,'p_g2':p_g2,'p_g3':p_g3,'p_g4':p_g4,
# 'p_u1':p_u1,'p_u2':p_u2,'p_u3':p_u3,'p_u4':p_u4,
# 'p_h1':p_h1,'p_h2':p_h2,'p_h3':p_h3,'p_h4':p_h4,
# }

# u_1,u_2 = 1.0,2.0

# u_dict = {
# 'u_1':1.0,'u_2':2.0
# }

# h_dict ={
# 'z_1':z[0],'z_2':z[1]
# } 



# import dae1

# model = dae1.model()
# model.ini({},1)
# #model.report_x()
# #model.report_y()
# #model.report_u()
# #model.report_z()
# #model.report_params()
# #model.jac_ini

# x_1,x_2,x_3 = model.get_mvalue(['x_1','x_2','x_3'])
# y_1,y_2,y_3,y_4 = model.get_mvalue(['y_1','y_2','y_3','y_4'])

# x = np.array([[x_1],   
#               [x_2],   
#               [x_3]]).reshape(3,)

# y = np.array([[y_1],   
#               [y_2],   
#               [y_3],   
#               [y_4]]).reshape(4,)

# u = np.array([[u_1],   
#               [u_2]]).reshape(2,)


# p = np.array([p_f1,p_f2,p_f3,p_f4,
#               p_g1,p_g2,p_g3,p_g4,
#               p_u1,p_u2,p_u3,p_u4,
#               p_h1,p_h2,p_h3,p_h4]).reshape(16,)

# Dt = 1e-3

# F_x =  np.array([[-p_f1, 0, y_1], [0, 0, -p_f2], [0, -p_f3, 0]])
# F_y =  np.array([[p_f4 + x_3, 0, 0, 0], [0, 0, 0, 0], [u_2, 0, 0, 0]])
# F_u =  np.array([[p_u1, 0], [0, 0], [0, y_1]])
# G_x =  np.array([[0, 0, 0], [0, y_1, 0], [0, 0, y_2], [0, y_2, 0]])
# G_y =  np.array([[0, p_g1, 0, 0], [p_g2 + x_2, 0, 0, 0], [0, x_3, p_g3, 0], [0, x_2, 0, u_1]])
# G_u =  np.array([[p_u2, 0], [0, 0], [0, 0], [y_4, p_u3]])

# jac_ini = np.array([[-p_f1, 0, y_1, p_f4 + x_3, 0, 0, 0], 
#                     [0, 0, -p_f2, 0, 0, 0, 0], 
#                     [0, -p_f3, 0, u_2, 0, 0, 0], 
#                     [0, 0, 0, 0, p_g1, 0, 0], 
#                     [0, y_1, 0, p_g2 + x_2, 0, 0, 0], 
#                     [0, 0, y_2, 0, x_3, p_g3, 0], 
#                     [0, y_2, 0, 0, x_2, 0, u_1]])


# assert (np.max(np.abs(jac_ini-model.jac_ini))) < 1e-16

# from dae1 import de_jac_ini_eval,sp_jac_ini_eval

# N_x = F_x.shape[0]
# N_y = G_y.shape[0]
# N_xy = N_x+N_y
# de_jac_ini = np.zeros((N_xy,N_xy))
# de_jac_ini_eval(de_jac_ini,x,y,u,p,Dt)
# #print(np.max(np.abs(jac_ini-de_jac_ini)))

# from scipy.sparse import csr_array
# from scipy.sparse import csr_matrix

# sp_jac_ini = csr_matrix((np.zeros(len(jac_ini_data)),jac_ini_indices,jac_ini_indptr),shape=jac_ini_shape)
# sp_jac_ini_eval(sp_jac_ini.data,x,y,u,p,Dt)

# assert (np.max(np.abs(jac_ini-sp_jac_ini))) < 1e-16
# #print(de_jac_ini)
# #print(sp_jac_ini.toarray())
# #print(sp_jac_ini.indices)
# #print(sp_jac_ini.indptr)
# #print(sp_jac_ini.nnz)
# #
# #print(jac_ini_indices,jac_ini_indptr)

# jac_trap =  np.array([[0.5*Dt*p_f1 + 1, 0, -0.5*Dt*y_1, -0.5*Dt*(p_f4 + x_3), 0, 0, 0], 
#                       [0, 1, 0.5*Dt*p_f2, 0, 0, 0, 0], 
#                       [0, 0.5*Dt*p_f3, 1, -0.5*Dt*u_2, 0, 0, 0], 
#                       [0, 0, 0, 0, p_g1, 0, 0], 
#                       [0, y_1, 0, p_g2 + x_2, 0, 0, 0], 
#                       [0, 0, y_2, 0, x_3, p_g3, 0], 
#                       [0, y_2, 0, 0, x_2, 0, u_1]])


# -*- coding: utf-8 -*-
"""
Created on Wed Dec 15 23:38:58 2021

@author: jmmau
"""

import pytest
import numpy as np
from pydae.bmapu import bmapu_builder
 
 
def test_vsc_l_pq_builder():
    data = {
    "system":{"name":"smib_vsc_l","S_base":100e6, "K_p_agc":0.0,"K_i_agc":0.0,"K_xif":0.01},       
    "buses":[{"name":"1", "P_W":0.0,"Q_var":0.0,"U_kV":20.0},
            {"name":"2", "P_W":0.0,"Q_var":0.0,"U_kV":20.0}
            ],
    "lines":[{"bus_j":"1", "bus_k":"2", "X_pu":0.05,"R_pu":0.01,"Bs_pu":1e-6,"S_mva":100.0}],
    "vscs": [{"bus":"1","type":"vsc_l","S_n":1e6,"F_n":50.0,"X_s":0.1,"R_s":0.01,
            "ctrl":{"type":"pq","p_s_ref":0.2,"q_s_ref":0.1}}],
    "genapes":[
        {"bus":"2","S_n":100e6,"F_n":50.0,"X_v":0.001,"R_v":0.0,"K_delta":0.001,"K_alpha":1e-6}]
    }
    
    grid = bmapu_builder.bmapu(data)
    #grid.checker()
    grid.uz_jacs = True
    grid.verbose = False
    grid.build('smib_vsc_l_pq')


def test_vsc_l():

    import smib_vsc_l_pq

    model = smib_vsc_l_pq.model()
    model.ini({},'xy_0.json')

    X_s = model.get_value('X_s_1')
    R_s = model.get_value('R_s_1')
    V_s = model.get_value('V_1')
    theta_s = model.get_value('theta_1')
    i_sr = model.get_value('i_sr_1')
    i_si = model.get_value('i_si_1')

    v_s = V_s*np.exp(1j*theta_s)
    i_s = i_sr + 1j*i_si

    v_sr = v_s.real   # X_m*cos(X_a)
    v_si = v_s.imag   # X_m*sin(X_a)

    i_sr = i_s.real   # X_m*cos(X_a)
    i_si = i_s.imag   # X_m*sin(X_a)

    # test the impedance are propoperly implemented
    v_ti_exp =  R_s*i_si + X_s*i_sr + v_si  
    v_tr_exp =  R_s*i_sr - X_s*i_si + v_sr 


    m = model.get_value('m_1')
    theta_t = model.get_value('theta_t_1')
    v_dc = model.get_value('v_dc_1')

    v_t_m = m*v_dc
    v_tr =  v_t_m*np.cos(theta_t)
    v_ti =  v_t_m*np.sin(theta_t)

    tol = 1e-8 

    assert np.abs(v_tr_exp - v_tr) < tol  # v_tr
    assert np.abs(v_ti_exp - v_ti) < tol  # v_ti

    # check powers are computed properly
    s_s_exp = v_s*np.conj(i_s)

    p_s_ri = i_sr*v_sr + i_si*v_si
    q_s_ri = i_sr*v_si - i_si*v_sr

    assert np.abs(s_s_exp.real - p_s_ri) < tol  # p_s
    assert np.abs(s_s_exp.imag - q_s_ri) < tol  # q_s


def test_vsc_l_pq():

    import smib_vsc_l_pq

    model = smib_vsc_l_pq.model()
    model.ini({'p_s_ref_1':0.5, 'q_s_ref_1':0.3},'xy_0.json')

    p_s_ref_1 = model.get_value('p_s_ref_1')
    q_s_ref_1 = model.get_value('q_s_ref_1')
    p_s_1 = model.get_value('p_s_1')
    q_s_1 = model.get_value('q_s_1')

    tol = 1e-8
    assert np.abs(p_s_ref_1 - p_s_1) < tol  # p_s
    assert np.abs(q_s_ref_1 - q_s_1) < tol  # q_s

    #delta = -0.1
    #v_sd = v_si * np.cos(delta) - v_sr * np.sin(delta)   
    #v_sq = v_si * np.sin(delta) + v_sr * np.cos(delta)
    #i_sd = i_si * np.cos(delta) - i_sr * np.sin(delta)   
    #i_sq = i_si * np.sin(delta) + i_sr * np.cos(delta)
#
    #v_td =  R_s*i_sd + X_s*i_sq + v_sd  
    #v_tq =  R_s*i_sq - X_s*i_sd + v_sq 
#
    #v_ti =  v_td * np.cos(delta) + v_tq * np.sin(delta)   
    #v_tr = -v_td * np.sin(delta) + v_tq * np.cos(delta) 
#
    #print(f'v_tr = {v_tr:0.3f}, v_ti = {v_ti:0.3f}')
#
    #
#
    #p_s_ri = i_sr*v_sr + i_si*v_si
    #q_s_ri = i_sr*v_si - i_si*v_sr
#
    #p_s_dq = i_sd*v_sd + i_sq*v_sq
    #q_s_dq = i_sq*v_sd - i_sd*v_sq
#
    #print(f'p_s =    {s_s.real:0.3f},    q_s = {s_s.imag:0.3f}')
    #print(f'p_s_ri = {p_s_ri:0.3f}, q_s_ri = {q_s_ri:0.3f}')
    #print(f'p_s_dq = {p_s_dq:0.3f}, q_s_dq = {q_s_dq:0.3f}')


if __name__ == "__main__":
    
    #test_pendulum_builder()
    pass 
    #test_leon_vsg_ll()


# -*- coding: utf-8 -*-
"""
Created on Thu August 10 23:52:55 2022

@author: jmmauricio
"""


import sympy as sym


def ntsst1(dae,syn_data,name):
    '''

    .. table:: Constants
        :widths: auto

    Example:
    
    ``"avr":{"type":"ntsst1","K_a":200.0,"T_c":1.0,"T_b":10.0,"v_ref":1.0},``

    '''

    T_r=0.01

    avr_data = syn_data['avr']
    
    v_t = sym.Symbol(f"V_{name}", real=True)   
    v_c = sym.Symbol(f"v_c_{name}", real=True)  
    x_cb  = sym.Symbol(f"x_cb_{name}", real=True)
    xi_v  = sym.Symbol(f"xi_v_{name}", real=True)
    v_f = sym.Symbol(f"v_f_{name}", real=True)  
    T_b = sym.Symbol(f"T_b_{name}", real=True) 
    T_c = sym.Symbol(f"T_c_{name}", real=True) 

    K_a = sym.Symbol(f"K_a_{name}", real=True)
    K_ai = sym.Symbol(f"K_ai_{name}", real=True)

    v_ref = sym.Symbol(f"v_ref_{name}", real=True) 
    v_pss = sym.Symbol(f"v_pss_{name}", real=True) 

    v_s = v_pss # v_oel and v_uel are not considered
    v_ini = K_ai*xi_v

    v_c = v_t # no droop is considered
    v_1 = v_ref - v_c + v_s + v_ini # v_ini is added in pydae to force V = v_ref in the initialization
    
    epsilon_v = v_ref - v_c 
    
    dx_cb = (v_1 - x_cb)/T_b;  
    z_cb  = (v_1 - x_cb)*T_c/T_b + x_cb 
    dxi_v = epsilon_v  # this integrator is added in pydae to force V = v_ref in the initialization

    g_v_f  =   K_a*z_cb - v_f 
    
    dae['f'] += [dx_cb,dxi_v]
    dae['x'] += [ x_cb, xi_v]
    dae['g'] += [g_v_f]
    dae['y_ini'] += [v_f] 
    dae['y_run'] += [v_f]  

    dae['params_dict'].update({str(K_a):avr_data['K_a']})
    dae['params_dict'].update({str(K_ai):1e-6})
    dae['params_dict'].update({str(T_c):avr_data['T_c']})  
    dae['params_dict'].update({str(T_b):avr_data['T_b']})  

    dae['u_ini_dict'].update({str(v_ref):avr_data['v_ref']})
    dae['u_run_dict'].update({str(v_ref):avr_data['v_ref']})

    dae['u_run_dict'].update({str(v_pss):0.0})
    dae['u_ini_dict'].update({str(v_pss):0.0})

    dae['xy_0_dict'].update({str(v_f):2.0})
    dae['xy_0_dict'].update({str(xi_v):10.0})
    dae['xy_0_dict'].update({str(x_cb):2.0})

    dae['h_dict'].update({str(v_ref):v_ref})  


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Apr 14 11:53:28 2018

@author: jmmauricio

todo:
    - save matrices 
    - add dense matrices
    - add (i,j) to lists
    - add f_ini and f_run

"""
import numpy as np
import sympy as sym
from collections import deque 
import pkgutil
import re
from sympy.matrices.sparsetools import _doktocsr
from sympy import SparseMatrix
from sympy.codegen.ast import Assignment
from scipy.sparse import csr_matrix,save_npz
import os
import cffi
import time
import logging




class builder():   

    def __init__(self,sys,verbose=False):

        logging.basicConfig(format='%(asctime)s %(message)s', level=logging.ERROR)

        self.verbose = verbose
        self.sys = sys
        self.jac_num = True
        self.name = self.sys['name']
        self.save_sources = False
        self.string_u2z = ''
        self.u2z = '\#'
        self.inirun = True


        if not os.path.exists('build'):
            os.makedirs('build')

        self.matrices_folder = 'build'

        self.f_ini_list = []
        self.f_run_list = []
        self.g_ini_list = []
        self.g_run_list = []
        self.h_list = []

        self.jac_ini_list = []
        self.jac_run_list = []
        self.jac_trap_list = []

        self.Fu_list = []
        self.Gu_list = []
        self.Hx_list = []
        self.Hy_list = []
        self.Hu_list = []

        if not 'uz_jacs' in sys:
            self.uz_jacs = False
            self.string_u2z = '\n'
            self.u2z_comment = '#'
        else:
            self.uz_jacs = sys['uz_jacs'] 
            self.u2z_comment = ''

    def dict2system(self):
        '''
        Converts input dictionary of DAE system to sympy matrices of the DAE system


        Parameters
        ----------
        sys : dict
            {'name':sys_name,
                'params_dict':params_dict,
                'f_list':f_list,
                'g_list':g_list,
                'x_list':x_list,
                'y_ini_list':y_ini_list,
                'y_run_list':y_run_list,
                'u_ini_dict':u_ini_dict,
                'u_run_dict':u_run_dict,
                'h_dict':h_dict}

        Returns
        -------
        sys : TYPE
            DESCRIPTION.

        '''

        sys = self.sys
        
        t_0 = time.time()
        logging.debug('check_system')
        self.check_system()
        
        # from dict to sympy matrices and vectors
        f = sym.Matrix(sys['f_list']).T
        g = sym.Matrix(sys['g_list']).T
        x = sym.Matrix(sys['x_list']).T
        y_ini = sym.Matrix(sys['y_ini_list']).T
        y_run = sym.Matrix(sys['y_run_list']).T
        u_ini = sym.Matrix(list(sys['u_ini_dict'].keys()), real=True)
        u_run_list = [sym.Symbol(item,real=True) for item in list(sys['u_run_dict'].keys())]
        u_run = sym.Matrix(u_run_list).T 
        h =  sym.Matrix(list(sys['h_dict'].values())).T   
        

        # jacobians computation
        logging.debug('computing jacobians Fx_run,Fy_run')
        Fx = sym_jac(f,x) 
        Fy_run = sym_jac(f,y_run)
        logging.debug('computing jacobians Gx_run,Gy_run')
        Gx = sym_jac(g,x)

        Gy_run = sym_jac(g,y_run)
        logging.debug('computing jacobians Fu_run,Gu_run')
        Fu_run = sym_jac(f,u_run)
        Gu_run = sym_jac(g,u_run)

        if self.inirun: # do not compute ini jacobians if ini system is equal to run system
            logging.debug('computing jacobians Fx_ini,Fy_ini')
            Fy_ini = sym_jac(f,y_ini)
            logging.debug('computing jacobians Gy_ini')
            Gy_ini = sym_jac(g,y_ini) 
        else:
            Fy_ini = Fy_run
            Gy_ini = Gy_run

        logging.debug('computing jacobians Hx_run,Hy_run,Hu_run')
        Hx_run = sym_jac(h,x) 
        Hy_run = sym_jac(h,y_run)
        Hu_run = sym_jac(h,u_run)

        # dictionary update
        sys['f']= f
        sys['g']= g 
        sys['x']= x
        sys['y_ini']= y_ini  
        sys['y_run']= y_run      
        sys['u_ini']= u_ini 
        sys['u_run']= u_run
        sys['h']= h 
        
        sys['Fx'] = Fx
        sys['Fy_run'] = Fy_run
        sys['Gy_run'] = Gy_run

        sys['Fx'] = Fx
        sys['Fy_ini'] = Fy_ini
        sys['Gx'] = Gx
        sys['Gy_ini'] = Gy_ini
        
        sys['Fu_run'] = Fu_run
        sys['Gu_run'] = Gu_run

        sys['Hx_run'] = Hx_run
        sys['Hy_run'] = Hy_run
        sys['Hu_run'] = Hu_run
        
        Fx_ini_rows,Fx_ini_cols = np.nonzero(Fx)
        Fy_ini_rows,Fy_ini_cols = np.nonzero(Fy_ini)
        Gx_ini_rows,Gx_ini_cols = np.nonzero(Gx)
        Gy_ini_rows,Gy_ini_cols = np.nonzero(Gy_ini)
        
        sys['Fx_ini_rows'] = Fx_ini_rows 
        sys['Fx_ini_cols'] = Fx_ini_cols 
        sys['Fy_ini_rows'] = Fy_ini_rows 
        sys['Fy_ini_cols'] = Fy_ini_cols 
        sys['Gx_ini_rows'] = Gx_ini_rows 
        sys['Gx_ini_cols'] = Gx_ini_cols 
        sys['Gy_ini_rows'] = Gy_ini_rows 
        sys['Gy_ini_cols'] = Gy_ini_cols 

        Fx_run_rows,Fx_run_cols = np.nonzero(Fx)
        Fy_run_rows,Fy_run_cols = np.nonzero(Fy_run)
        Gx_run_rows,Gx_run_cols = np.nonzero(Gx)
        Gy_run_rows,Gy_run_cols = np.nonzero(Gy_run)
        
        sys['Fx_run_rows'] = Fx_run_rows 
        sys['Fx_run_cols'] = Fx_run_cols 
        sys['Fy_run_rows'] = Fy_run_rows 
        sys['Fy_run_cols'] = Fy_run_cols 
        sys['Gx_run_rows'] = Gx_run_rows 
        sys['Gx_run_cols'] = Gx_run_cols 
        sys['Gy_run_rows'] = Gy_run_rows 
        sys['Gy_run_cols'] = Gy_run_cols 
        
        sys['N_x'] = len(x)
        sys['N_y'] = len(y_run)
        sys['N_u'] = len(u_run)
        sys['N_z'] = len(h)

        logging.debug('end checking and computing jacobians')

        self.sys = sys

    def check_system(self):
        
        if len(self.sys['f_list']) == 0:
            print('system without dynamic equations, adding dummy dynamic equation')
            x_dummy,u_dummy = sym.symbols('x_dummy,u_dummy')
            self.sys['x_list'] = ['x_dummy']
            self.sys['f_list'] = [u_dummy-x_dummy]
            self.sys['u_ini_dict'].update({'u_dummy':1.0})
            self.sys['u_run_dict'].update({'u_dummy':1.0})

        if len(self.sys['g_list']) == 0:
            print('system without algebraic equations, adding dummy algebraic equation')
            y_dummy,u_dummy = sym.symbols('y_dummy,u_dummy')
            
            self.sys['g_list'] = [u_dummy-y_dummy]
            self.sys['y_ini_list'] = ['y_dummy']
            self.sys['y_run_list'] = ['y_dummy']
            self.sys['u_ini_dict'].update({'u_dummy':1.0})
            self.sys['u_run_dict'].update({'u_dummy':1.0})
            
        y_ini_set = set(self.sys['y_ini_list'])
        contains_duplicates = len(self.sys['y_ini_list']) != len(y_ini_set)
        if contains_duplicates:
            print('error: y_ini contains duplicates')

        y_set = set(self.sys['y_run_list'])
        contains_duplicates = len(self.sys['y_run_list']) != len(y_set)
        if contains_duplicates:
            print('error: y_run contains duplicates')

        if sys['y_run_list'] == ['y_ini_list']:
            self.inirun = False

    def functions(self):
        
        logging.debug('f_ini symbolic to c and xyup')
        f = self.sys['f']
        for item in f:
            self.f_ini_list += [{'sym':item}]
        sym2c(self.f_ini_list)
        sym2xyup(self.sys,self.f_ini_list,'ini')
        logging.debug('end f_ini symbolic to c and xyup')

        logging.debug('f_run symbolic to c and xyup')
        f = self.sys['f']
        for item in f:
            self.f_run_list += [{'sym':item}]
        sym2c(self.f_run_list)
        sym2xyup(self.sys,self.f_run_list,'run')
        logging.debug('end f symbolic to c and xyup')

        logging.debug('g_ini symbolic to c and xyup')
        for item in self.sys['g']:
            self.g_ini_list += [{'sym':item}]
        sym2c(self.g_ini_list)
        sym2xyup(self.sys,self.g_ini_list,'ini')
        logging.debug('end g_ini symbolic to c and xyup')

        logging.debug('g_run symbolic to c and xyup')
        for item in self.sys['g']:
            self.g_run_list += [{'sym':item}]
        sym2c(self.g_run_list)
        sym2xyup(self.sys,self.g_run_list,'run')
        logging.debug('end g_run symbolic to c and xyup')

        logging.debug('h symbolic to c and xyup')
        for item in self.sys['h']:
            self.h_list += [{'sym':item}]
        sym2c(self.h_list)
        sym2xyup(self.sys,self.h_list,'run')
        logging.debug('end h symbolic to c and xyup')

    def jacobians(self):
        '''
        

        Parameters
        ----------
        sys : dict
            

        Returns
        -------
        sys : TYPE
            DESCRIPTION.

        '''
                
        N_x = self.sys['N_x']
    
        Fx_run = self.sys['Fx']
        Fy_run = self.sys['Fy_run']
        Gx_run = self.sys['Gx']
        Gy_run = self.sys['Gy_run']
        
        Fx_ini = self.sys['Fx']
        Fy_ini = self.sys['Fy_ini']
        Gx_ini = self.sys['Gx']
        Gy_ini = self.sys['Gy_ini']

        Fu_run = self.sys['Fu_run']
        Gu_run = self.sys['Gu_run']  

        Hx_run = self.sys['Hx_run']
        Hy_run = self.sys['Hy_run']
        Hu_run = self.sys['Hu_run']

        logging.debug('computing jac_ini')
        jac_ini = sym.Matrix([[Fx_ini,Fy_ini],[Gx_ini,Gy_ini]]) 
        
        logging.debug('computing jac_run')
        jac_run = sym.Matrix([[Fx_run,Fy_run],[Gx_run,Gy_run]]) 
        
        logging.debug('computing jac_trap')
        eye = sym.eye(N_x, real=True)
        Dt = sym.Symbol('Dt',real=True)
        jac_trap = sym.Matrix([[eye - 0.5*Dt*Fx_run, -0.5*Dt*Fy_run],[Gx_run,Gy_run]])    

        self.sys['jac_ini']  = jac_ini
        self.sys['jac_run']  = jac_run   
        self.sys['jac_trap'] = jac_trap
        
        logging.debug('end of large jacobians computation')

        ## jac_ini
        self.jac_ini_sp = _doktocsr(SparseMatrix(jac_ini))
        self.sys['sp_jac_ini_list'] = self.jac_ini_sp
        ij_indices,dense_indices = spidx2ij(self.jac_ini_sp)
        logging.debug('jac_ini symbolic to c')
        for it,item in enumerate(self.jac_ini_sp[0]):
            self.jac_ini_list += [{'sym':item,'ij':ij_indices[it],'de_idx':dense_indices[it]}]

        sym2c(self.jac_ini_list)
        logging.debug('end jac_ini symbolic to c')

        logging.debug('jac_ini c to c xyup')
        sym2xyup(self.sys,self.jac_ini_list,'ini')
        logging.debug('end jac_ini c to c xyup')

        ## jac_run
        self.jac_run_sp = _doktocsr(SparseMatrix(jac_run))
        self.sys['sp_jac_run_list'] = self.jac_run_sp
        ij_indices,dense_indices = spidx2ij(self.jac_run_sp)

        logging.debug('jac_run symbolic to c')
        for it,item in enumerate(self.jac_run_sp[0]):
            self.jac_run_list += [{'sym':item,'ij':ij_indices[it],'de_idx':dense_indices[it]}]

        sym2c(self.jac_run_list)
        logging.debug('end jac_run symbolic to c')

        logging.debug('jac_run c to c xyup')
        sym2xyup(self.sys,self.jac_run_list,'ini')
        logging.debug('end jac_run c to c xyup')

        ## jac_trap
        self.jac_trap_sp = _doktocsr(SparseMatrix(jac_trap))
        self.sys['sp_jac_trap_list'] = self.jac_trap_sp
        ij_indices,dense_indices = spidx2ij(self.jac_trap_sp)

        logging.debug('jac_trap symbolic to c')
        for it,item in enumerate(self.jac_trap_sp[0]):
            self.jac_trap_list += [{'sym':item,'ij':ij_indices[it],'de_idx':dense_indices[it]}]

        sym2c(self.jac_trap_list)
        logging.debug('end jac_trap symbolic to c')

        logging.debug('jac_trap c to c xyup')
        sym2xyup(self.sys,self.jac_trap_list,'run')
        logging.debug('end jac_trap c to c xyup')  

        ## Fu_run
        self.Fu_run_sp = _doktocsr(SparseMatrix(Fu_run))
        for item in self.Fu_run_sp[0]:
            self.Fu_list += [{'sym':item}]
        sym2c(self.Fu_list)
        sym2xyup(self.sys,self.Fu_list,'run')

        ## Gu_run
        self.Gu_run_sp = _doktocsr(SparseMatrix(Gu_run))
        for item in self.Gu_run_sp[0]:
            self.Gu_list += [{'sym':item}]
        sym2c(self.Gu_list)
        sym2xyup(self.sys,self.Gu_list,'run')

        ## Hx_run
        self.Hx_run_sp = _doktocsr(SparseMatrix(Hx_run))
        for item in self.Hx_run_sp[0]:
            self.Hx_list += [{'sym':item}]
        sym2c(self.Hx_list)
        sym2xyup(self.sys,self.Hx_list,'run')

        ## Hy_run
        self.Hy_run_sp = _doktocsr(SparseMatrix(Hy_run))
        for item in self.Hy_run_sp[0]:
            self.Hy_list += [{'sym':item}]
        sym2c(self.Hy_list)
        sym2xyup(self.sys,self.Hy_list,'run')

        ## Hu_run
        self.Hu_run_sp = _doktocsr(SparseMatrix(Hu_run))
        for item in self.Hu_run_sp[0]:
            self.Hu_list += [{'sym':item}]
        sym2c(self.Hu_list)
        sym2xyup(self.sys,self.Hu_list,'run')

    def check_system(self):
        sys = self.sys
        
        if len(sys['f_list']) == 0:
            print('system without dynamic equations, adding dummy dynamic equation')
            x_dummy,u_dummy = sym.symbols('x_dummy,u_dummy')
            sys['x_list'] = ['x_dummy']
            sys['f_list'] = [u_dummy-x_dummy]
            sys['u_ini_dict'].update({'u_dummy':1.0})
            sys['u_run_dict'].update({'u_dummy':1.0})

        if len(sys['g_list']) == 0:
            print('system without algebraic equations, adding dummy algebraic equation')
            y_dummy,u_dummy = sym.symbols('y_dummy,u_dummy')
            
            sys['g_list'] = [u_dummy-y_dummy]
            sys['y_ini_list'] = ['y_dummy']
            sys['y_run_list'] = ['y_dummy']
            sys['u_ini_dict'].update({'u_dummy':1.0})
            sys['u_run_dict'].update({'u_dummy':1.0})
            
        y_ini_set = set(sys['y_ini_list'])
        contains_duplicates = len(sys['y_ini_list']) != len(y_ini_set)
        if contains_duplicates:
            print('error: y_ini contains duplicates')

        y_set = set(sys['y_run_list'])
        contains_duplicates = len(sys['y_run_list']) != len(y_set)
        if contains_duplicates:
            print('error: y_run contains duplicates')

        if sys['y_run_list'] == ['y_ini_list']:
            self.inirun = False
  
    def cwrite(self): 

        defs = ''
        source = '' 

        defs += f'void f_ini_eval(double *data,double *x,double *y,double *u,double *p,double Dt);\n'
        source += f'void f_ini_eval(double *data,double *x,double *y,double *u,double *p,double Dt)' + '{' +'\n'*2
        for it,item in enumerate(self.f_ini_list):
            source += f"data[{it}] = {item['xyup']}; \n"
        source += '\n}\n\n'

        defs += f'void f_run_eval(double *data,double *x,double *y,double *u,double *p,double Dt);\n'
        source += f'void f_run_eval(double *data,double *x,double *y,double *u,double *p,double Dt)' + '{' +'\n'*2
        for it,item in enumerate(self.f_run_list):
            source += f"data[{it}] = {item['xyup']}; \n"
        source += '\n}\n\n'

        defs += f'void g_ini_eval(double *data,double *x,double *y,double *u,double *p,double Dt);\n'
        source += f'void g_ini_eval(double *data,double *x,double *y,double *u,double *p,double Dt)' + '{' +'\n'*2
        for it,item in enumerate(self.g_ini_list):
            source += f"data[{it}] = {item['xyup']}; \n"
        source += '\n}\n\n'

        defs += f'void g_run_eval(double *data,double *x,double *y,double *u,double *p,double Dt);\n'
        source += f'void g_run_eval(double *data,double *x,double *y,double *u,double *p,double Dt)' + '{' +'\n'*2
        for it,item in enumerate(self.g_run_list):
            source += f"data[{it}] = {item['xyup']}; \n"
        source += '\n}\n\n'

        defs += f'void h_eval(double *data,double *x,double *y,double *u,double *p,double Dt);\n'
        source += f'void h_eval(double *data,double *x,double *y,double *u,double *p,double Dt)' + '{' +'\n'*2
        for it,item in enumerate(self.h_list):
            source += f"data[{it}] = {item['xyup']}; \n"
        source += '\n}\n\n'

        # jac_ini dense
        defs += f'void de_jac_ini_up_eval(double *data,double *x,double *y,double *u,double *p,double Dt);\n'
        source += f'void de_jac_ini_up_eval(double *data,double *x,double *y,double *u,double *p,double Dt)' + '{' +'\n'*2
        for it,item in enumerate(self.jac_ini_list):
            if item['tipo'] == 'up':
                source += f"data[{item['de_idx']}] = {item['xyup']}; \n"
        source += '\n}\n\n'

        defs += f'void de_jac_ini_xy_eval(double *data,double *x,double *y,double *u,double *p,double Dt);\n'
        source += f'void de_jac_ini_xy_eval(double *data,double *x,double *y,double *u,double *p,double Dt)' + '{' +'\n'*2
        for it,item in enumerate(self.jac_ini_list):
            if item['tipo'] == 'xy':
                source += f"data[{item['de_idx']}] = {item['xyup']}; \n"
        source += '\n}\n\n'

        defs += f'void de_jac_ini_num_eval(double *data,double *x,double *y,double *u,double *p,double Dt);\n'
        source += f'void de_jac_ini_num_eval(double *data,double *x,double *y,double *u,double *p,double Dt)' + '{' +'\n'*2
        for it,item in enumerate(self.jac_ini_list):
            if item['tipo'] == 'num':   
                source += f"data[{item['de_idx']}] = {item['xyup']}; \n"
        source += '\n}\n\n'

        # jac_ini sparse
        defs += f'void sp_jac_ini_up_eval(double *data,double *x,double *y,double *u,double *p,double Dt);\n'
        source += f'void sp_jac_ini_up_eval(double *data,double *x,double *y,double *u,double *p,double Dt)' + '{' +'\n'*2
        for it,item in enumerate(self.jac_ini_list):
            if item['tipo'] == 'up':
                source += f"data[{it}] = {item['xyup']}; \n"
        source += '\n}\n\n'

        defs += f'void sp_jac_ini_xy_eval(double *data,double *x,double *y,double *u,double *p,double Dt);\n'
        source += f'void sp_jac_ini_xy_eval(double *data,double *x,double *y,double *u,double *p,double Dt)' + '{' +'\n'*2
        for it,item in enumerate(self.jac_ini_list):
            if item['tipo'] == 'xy':
                source += f"data[{it}] = {item['xyup']}; \n"
        source += '\n}\n\n'

        defs += f'void sp_jac_ini_num_eval(double *data,double *x,double *y,double *u,double *p,double Dt);\n'
        source += f'void sp_jac_ini_num_eval(double *data,double *x,double *y,double *u,double *p,double Dt)' + '{' +'\n'*2
        for it,item in enumerate(self.jac_ini_list):
            if item['tipo'] == 'num':
                source += f"data[{it}] = {item['xyup']}; \n"
        source += '\n}\n\n'

        # jac_run dense
        defs += f'void de_jac_run_up_eval(double *data,double *x,double *y,double *u,double *p,double Dt);\n'
        source += f'void de_jac_run_up_eval(double *data,double *x,double *y,double *u,double *p,double Dt)' + '{' +'\n'*2
        for it,item in enumerate(self.jac_run_list):
            if item['tipo'] == 'up':
                source += f"data[{item['de_idx']}] = {item['xyup']}; \n"
        source += '\n}\n\n'

        defs += f'void de_jac_run_xy_eval(double *data,double *x,double *y,double *u,double *p,double Dt);\n'
        source += f'void de_jac_run_xy_eval(double *data,double *x,double *y,double *u,double *p,double Dt)' + '{' +'\n'*2
        for it,item in enumerate(self.jac_run_list):
            if item['tipo'] == 'xy':
                source += f"data[{item['de_idx']}] = {item['xyup']}; \n"
        source += '\n}\n\n'

        defs += f'void de_jac_run_num_eval(double *data,double *x,double *y,double *u,double *p,double Dt);\n'
        source += f'void de_jac_run_num_eval(double *data,double *x,double *y,double *u,double *p,double Dt)' + '{' +'\n'*2
        for it,item in enumerate(self.jac_run_list):
            if item['tipo'] == 'num':
                source += f"data[{item['de_idx']}] = {item['xyup']}; \n"
        source += '\n}\n\n'

        # jac_run sparse
        defs += f'void sp_jac_run_up_eval(double *data,double *x,double *y,double *u,double *p,double Dt);\n'
        source += f'void sp_jac_run_up_eval(double *data,double *x,double *y,double *u,double *p,double Dt)' + '{' +'\n'*2
        for it,item in enumerate(self.jac_run_list):
            if item['tipo'] == 'up':
                source += f"data[{it}] = {item['xyup']}; \n"
        source += '\n}\n\n'

        defs += f'void sp_jac_run_xy_eval(double *data,double *x,double *y,double *u,double *p,double Dt);\n'
        source += f'void sp_jac_run_xy_eval(double *data,double *x,double *y,double *u,double *p,double Dt)' + '{' +'\n'*2
        for it,item in enumerate(self.jac_run_list):
            if item['tipo'] == 'xy':
                source += f"data[{it}] = {item['xyup']}; \n"
        source += '\n}\n\n'

        defs += f'void sp_jac_run_num_eval(double *data,double *x,double *y,double *u,double *p,double Dt);\n'
        source += f'void sp_jac_run_num_eval(double *data,double *x,double *y,double *u,double *p,double Dt)' + '{' +'\n'*2
        for it,item in enumerate(self.jac_run_list):
            if item['tipo'] == 'num':
                source += f"data[{it}] = {item['xyup']}; \n"
        source += '\n}\n\n'

        # jac_trap dense
        defs += f'void de_jac_trap_up_eval(double *data,double *x,double *y,double *u,double *p,double Dt);\n'
        source += f'void de_jac_trap_up_eval(double *data,double *x,double *y,double *u,double *p,double Dt)' + '{' +'\n'*2
        for it,item in enumerate(self.jac_trap_list):
            if item['tipo'] == 'up':
                source += f"data[{item['de_idx']}] = {item['xyup']}; \n"
        source += '\n}\n\n'

        defs += f'void de_jac_trap_xy_eval(double *data,double *x,double *y,double *u,double *p,double Dt);\n'
        source += f'void de_jac_trap_xy_eval(double *data,double *x,double *y,double *u,double *p,double Dt)' + '{' +'\n'*2
        for it,item in enumerate(self.jac_trap_list):
            if item['tipo'] == 'xy':
                source += f"data[{item['de_idx']}] = {item['xyup']}; \n"
        source += '\n}\n\n'

        defs += f'void de_jac_trap_num_eval(double *data,double *x,double *y,double *u,double *p,double Dt);\n'
        source += f'void de_jac_trap_num_eval(double *data,double *x,double *y,double *u,double *p,double Dt)' + '{' +'\n'*2
        for it,item in enumerate(self.jac_trap_list):
            if item['tipo'] == 'num':
                source += f"data[{item['de_idx']}] = {item['xyup']}; \n"
        source += '\n}\n\n'

        # jac_trap sparse
        defs += f'void sp_jac_trap_up_eval(double *data,double *x,double *y,double *u,double *p,double Dt);\n'
        source += f'void sp_jac_trap_up_eval(double *data,double *x,double *y,double *u,double *p,double Dt)' + '{' +'\n'*2
        for it,item in enumerate(self.jac_trap_list):
            if item['tipo'] == 'up':
                source += f"data[{it}] = {item['xyup']}; \n"
        source += '\n}\n\n'

        defs += f'void sp_jac_trap_xy_eval(double *data,double *x,double *y,double *u,double *p,double Dt);\n'
        source += f'void sp_jac_trap_xy_eval(double *data,double *x,double *y,double *u,double *p,double Dt)' + '{' +'\n'*2
        for it,item in enumerate(self.jac_trap_list):
            if item['tipo'] == 'xy':
                source += f"data[{it}] = {item['xyup']}; \n"
        source += '\n}\n\n'

        defs += f'void sp_jac_trap_num_eval(double *data,double *x,double *y,double *u,double *p,double Dt);\n'
        source += f'void sp_jac_trap_num_eval(double *data,double *x,double *y,double *u,double *p,double Dt)' + '{' +'\n'*2
        for it,item in enumerate(self.jac_trap_list):
            if item['tipo'] == 'num':
                source += f"data[{it}] = {item['xyup']}; \n"
        source += '\n}\n\n'

        if self.uz_jacs:
            self.string_u2z = 'sp_Fu_run_up_eval = jacs.lib.sp_Fu_run_up_eval\n'
            self.string_u2z+= 'sp_Gu_run_up_eval = jacs.lib.sp_Gu_run_up_eval\n'
            self.string_u2z+= 'sp_Hx_run_up_eval = jacs.lib.sp_Hx_run_up_eval\n'
            self.string_u2z+= 'sp_Hy_run_up_eval = jacs.lib.sp_Hy_run_up_eval\n'
            self.string_u2z+= 'sp_Hu_run_up_eval = jacs.lib.sp_Hu_run_up_eval\n'
            self.string_u2z+= 'sp_Fu_run_xy_eval = jacs.lib.sp_Fu_run_xy_eval\n'
            self.string_u2z+= 'sp_Gu_run_xy_eval = jacs.lib.sp_Gu_run_xy_eval\n'
            self.string_u2z+= 'sp_Hx_run_xy_eval = jacs.lib.sp_Hx_run_xy_eval\n'
            self.string_u2z+= 'sp_Hy_run_xy_eval = jacs.lib.sp_Hy_run_xy_eval\n'
            self.string_u2z+= 'sp_Hu_run_xy_eval = jacs.lib.sp_Hu_run_xy_eval\n'

            # Fu, Gu, Hx, Hy, Hu sparse
            matrices = [{'name':'Fu_run','lista':self.Fu_list,'matrix':self.Fu_run_sp,'de_sp':['sp'],'ini_run':['run']},
                        {'name':'Gu_run','lista':self.Gu_list,'matrix':self.Gu_run_sp,'de_sp':['sp'],'ini_run':['run']},
                        {'name':'Hx_run','lista':self.Hx_list,'matrix':self.Hx_run_sp,'de_sp':['sp'],'ini_run':['run']},
                        {'name':'Hy_run','lista':self.Hy_list,'matrix':self.Hy_run_sp,'de_sp':['sp'],'ini_run':['run']},
                        {'name':'Hu_run','lista':self.Hu_list,'matrix':self.Hu_run_sp,'de_sp':['sp'],'ini_run':['run']},]

            for item in matrices:
                for de_sp in item['de_sp']:
                    for xyup in ['up','xy','num']:
                        defs += f'void {de_sp}_{item["name"]}_{xyup}_eval(double *data,double *x,double *y,double *u,double *p,double Dt);\n'
                        source += f'void {de_sp}_{item["name"]}_{xyup}_eval(double *data,double *x,double *y,double *u,double *p,double Dt)' + '{' +'\n'*2
                        for it,item2 in enumerate(item["lista"]):
                            if item2['tipo'] == xyup:
                                source += f"data[{it}] = {item2['xyup']}; \n"
                        source += '\n}\n\n' 
                matrix = csr_matrix((np.zeros(len(item['matrix'][0])),item['matrix'][1],item['matrix'][2]),shape=item['matrix'][3])
                save_npz( f"./{self.matrices_folder}/{self.sys['name']}_{item['name']}_num.npz", matrix, compressed=True)

        sp_jac_ini_num_matrix = csr_matrix((np.zeros(len(self.jac_ini_sp[0])),self.jac_ini_sp[1],self.jac_ini_sp[2]),shape=self.jac_ini_sp[3])
        sp_jac_run_num_matrix = csr_matrix((np.zeros(len(self.jac_run_sp[0])),self.jac_run_sp[1],self.jac_run_sp[2]),shape=self.jac_run_sp[3])
        sp_jac_trap_num_matrix = csr_matrix((np.zeros(len(self.jac_trap_sp[0])),self.jac_trap_sp[1],self.jac_trap_sp[2]),shape=self.jac_trap_sp[3])

        save_npz( f"./{self.matrices_folder}/{self.sys['name']}_sp_jac_ini_num.npz", sp_jac_ini_num_matrix, compressed=True)
        save_npz( f"./{self.matrices_folder}/{self.sys['name']}_sp_jac_run_num.npz", sp_jac_run_num_matrix, compressed=True)
        save_npz( f"./{self.matrices_folder}/{self.sys['name']}_sp_jac_trap_num.npz",sp_jac_trap_num_matrix, compressed=True)


        self.defs = defs
        self.source = source

        with open(f'./build/defs_{self.name}_cffi.h', 'w') as fobj:
            fobj.write(self.defs)
        with open(f'./build/source_{self.name}_cffi.c', 'w') as fobj:
            fobj.write(self.source)


    def compile(self):
        
        logging.debug('start compiling module')
        ffi = cffi.FFI()
        name = self.name
            
        ffi.cdef(self.defs, override=True)
        ffi.set_source(module_name=f"{self.name}_cffi",source=self.source)
        ffi.compile()
        logging.debug('end compiling module')


    def template(self):
        sys = self.sys

        t_0 = time.time()
        
        name = sys['name']
        N_x = sys['N_x']
        N_y = sys['N_y']
        N_z = sys['N_z']
        N_u = sys['N_u']

        class_template = pkgutil.get_data(__name__, "templates/class_dae_cffi_template.py").decode().replace('\r\n','\n') 
        functions_template = pkgutil.get_data(__name__, "templates/functions_template.py").decode().replace('\r\n','\n') 
    #     solver_template = pkgutil.get_data(__name__, "templates/solver_template_v2.py").decode().replace('\r\n','\n') 

        class_template = class_template.replace('{name}',str(name))
        class_template = class_template.replace('{u2z_jacobians}',self.string_u2z)
        class_template = class_template.replace('{u2z_comment}',self.u2z_comment)

        class_template = class_template.replace('{N_x}',str(N_x)).replace('{N_y}',str(N_y)).replace('{N_z}',str(N_z))
        class_template = class_template.replace('{x_list}', str([str(item) for item in sys['x']]))
        class_template = class_template.replace('{y_run_list}', str([str(item) for item in sys['y_run']]))
        class_template = class_template.replace('{y_ini_list}', str([str(item) for item in sys['y_ini']]))
        class_template = class_template.replace('{params_list}', str([str(item) for item in sys['params_dict']]))
        class_template = class_template.replace('{params_values_list}', str([(sys['params_dict'][item]) for item in sys['params_dict']]))
        class_template = class_template.replace('{inputs_ini_list}', str([str(item) for item in sys['u_ini']]))
        class_template = class_template.replace('{inputs_run_list}', str([str(item) for item in sys['u_run']]))
        class_template = class_template.replace('{inputs_ini_values_list}', str([(sys['u_ini_dict'][item]) for item in sys['u_ini_dict']]))
        class_template = class_template.replace('{inputs_run_values_list}', str([(sys['u_run_dict'][item]) for item in sys['u_run_dict']]))
        class_template = class_template.replace('{outputs_list}', str([str(item) for item in sys['h_dict']]))

        if 'enviroment' in sys:
            class_template = class_template.replace('{enviroment_name}',str(sys['enviroment']))
            class_template = class_template.replace('{dae_file_mode}',"'enviroment'")
        elif 'colab' in sys:
            class_template = class_template.replace('{enviroment_name}',str(sys['colab']))
            class_template = class_template.replace('{dae_file_mode}',"'colab'")            
        else:
            class_template = class_template.replace('{enviroment_name}','no_enviroment')
            class_template = class_template.replace('{dae_file_mode}',"'local'")
            
        module = class_template
        
        module += '\n'*2
        module += self.spmatrix2def(sys['sp_jac_ini_list'],'sp_jac_ini')
        module += '\n'
        module += self.spmatrix2def(sys['sp_jac_run_list'],'sp_jac_run')
        module += '\n'
        module += self.spmatrix2def(sys['sp_jac_trap_list'],'sp_jac_trap')

        with open(f'{name}.py','w') as fobj:
            fobj.write(module)
    
        
    def spmatrix2def(self,spmatrix_list,matrix_name):
        
        string = ''
        string += f'def {matrix_name}_vectors():\n\n'
        string += f'{" "*4}{matrix_name}_ia = ' + str(spmatrix_list[1]) + '\n'
        string += f'{" "*4}{matrix_name}_ja = ' + str(spmatrix_list[2]) + '\n'    
        string += f'{" "*4}{matrix_name}_nia = ' + str(spmatrix_list[3][0])  + '\n'
        string += f'{" "*4}{matrix_name}_nja = ' + str(spmatrix_list[3][1])  + '\n'
        string += f'{" "*4}return {matrix_name}_ia, {matrix_name}_ja, {matrix_name}_nia, {matrix_name}_nja \n'
        
        return string   

def sym_jac(f,x):
    
    N_f = len(f)
    N_x = len(x)

    J = sym.MutableSparseMatrix(N_f, N_x, {})
    
    for irow in range(N_f):
        str_f = str(f[irow])
        for icol in range(N_x):
            if str(x[icol]) in str_f:
                J[irow,icol] = f[irow].diff(x[icol])

    return J

def spidx2ij(csr_matrix_list):

        data    = csr_matrix_list[0]
        indices = csr_matrix_list[1]
        indptr  = csr_matrix_list[2]
        shape   = csr_matrix_list[3]
        N_rows,N_cols = shape

        ij_indices = []
        dense_indices = []
        for i in range(shape[0]):
            row_col_start = indptr[i]
            row_col_end   = indptr[i+1]
            for j in indices[row_col_start:row_col_end]:
                ij_indices += [(i,j)]
                dense_indices += [i*N_cols+j]

        return ij_indices,dense_indices

def sym2c(full_list):
    '''
    Converts sympy symbolic expressions to c code
    '''

    for item in full_list:
        sym_expr = item['sym']
        c_expr = sym.ccode(sym_expr)
        item.update({'ccode':c_expr})


def sym2xyup(sys,full_list,inirun):
    '''
    Converts symbols to vectors x, y, u and p
    '''

    for element in full_list:
        string = element['ccode']
        i = 0
        for symbol in sys['x']:
            string = re.sub(f"\\b{symbol}\\b"  ,f'x[{i}]',string)
            i+=1

        i = 0
        for symbol in sys[f'y_{inirun}']:
            string = re.sub(f"\\b{symbol}\\b"  ,f'y[{i}]',string)
            i+=1

        i = 0
        for symbol in sys[f'u_{inirun}']:
            string = re.sub(f"\\b{symbol}\\b"  ,f'u[{i}]',string)
            i+=1

        i = 0
        for symbol in sys['params_dict']:
            string = re.sub(f"\\b{symbol}\\b"  ,f'p[{i}]',string)
            i+=1

        tipo = 'num'
        if 'x[' in string or 'y[' in string:
            tipo = 'xy' 
        elif 'p[' in string or 'u[' in string or 'Dt' in string:
            tipo = 'up' 
    
        element.update({'xyup':string,'tipo':tipo})






if __name__ == '__main__':
    from pydae.bmapu.bmapu_builder import  bmapu
    grid = bmapu(r"./smib.json")
    grid.construct('smib')

    b = builder(grid.sys_dict)
    b.dict2system()
    b.functions()
    b.jacobians()
    b.cwrite()
    #b.compile()

